"""Sign user out of Cognito and remove all Cookie Headers."""

import logging
from urllib.parse import urlencode

from shared import (
    create_error_html,
    extract_and_parse_cookies,
    generate_cookie_headers,
    get_config,
)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the signout event."""
    request = event["Records"][0]["cf"]["request"]
    domain_name = request["headers"]["host"][0]["value"]
    extracted = extract_and_parse_cookies(request["headers"], CONFIG["client_id"])

    if not extracted["id_token"]:
        response = {
            "body": create_error_html(
                "Signed out",
                "You are already signed out",
                f"https://{domain_name}{CONFIG['redirect_path_sign_out']}",
                "Proceed",
            ),
            "status": "200",
            "headers": {
                "content-type": [{"key": "Content-Type", "value": "text/html; charset=UTF-8"}],
                **CONFIG.get("cloud_front_headers", {}),
            },
        }
        return response

    tokens = {
        "id_token": extracted["id_token"],
        "access_token": extracted["access_token"],
        "refresh_token": extracted["refresh_token"],
    }
    query_string = {
        "logout_uri": f"https://{domain_name}{CONFIG['redirect_path_sign_out']}",
        "client_id": CONFIG["client_id"],
    }

    response = {
        "status": "307",
        "statusDescription": "Temporary Redirect",
        "headers": {
            # Redirect the user to logout
            "location": [
                {
                    "key": "location",
                    "value": "https://%s/logout?%s"
                    % (CONFIG.get("cognito_auth_domain"), urlencode(query_string)),
                }
            ],
            "set-cookie": generate_cookie_headers(
                "sign_out",
                CONFIG.get("client_id"),
                CONFIG.get("oauth_scopes"),
                tokens,
                domain_name,
                CONFIG.get("cookie_settings"),
            ),
            **CONFIG.get("cloud_front_headers", {}),
        },
    }
    return response



"""Runway config variables definition."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, cast

import yaml

from ....exceptions import VariablesFileNotFound
from ....utils import MutableMap
from ...models.runway import RunwayVariablesDefinitionModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


class RunwayVariablesDefinition(MutableMap):
    """Runway variables definition."""

    default_names = ["runway.variables.yml", "runway.variables.yaml"]

    # used to track persistent state on the class to only log the message once
    _has_notified_missing_file: ClassVar[bool] = False

    def __init__(self, data: RunwayVariablesDefinitionModel) -> None:
        """Instantiate class."""
        self._file_path = data.file_path
        self._sys_path = data.sys_path
        data = RunwayVariablesDefinitionModel(**{**data.model_dump(), **self.__load_file()})
        super().__init__(**data.model_dump(exclude={"file_path", "sys_path"}))

    def __load_file(self) -> dict[str, Any]:
        """Load a variables file."""
        if self._file_path:
            if self._file_path.is_file():
                return yaml.safe_load(self._file_path.read_text())
            raise VariablesFileNotFound(self._file_path.absolute())

        for name in self.default_names:
            test_path = self._sys_path / name
            LOGGER.debug("looking for variables file: %s", test_path)
            if test_path.is_file():
                LOGGER.verbose("found variables file: %s", test_path)
                return yaml.safe_load(test_path.read_text())

        if not self._has_notified_missing_file:
            LOGGER.info(
                "could not find %s in the current directory; continuing without a variables file",
                " or ".join(self.default_names),
            )
            self.__class__._has_notified_missing_file = True  # noqa: SLF001
        return {}

    @classmethod
    def parse_obj(cls: type[Self], obj: Any) -> Self:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        return cls(RunwayVariablesDefinitionModel.model_validate(obj))



"""Set package version."""

from __future__ import annotations

import logging

from ._logging import LogLevels, RunwayLogger  # noqa: F401

logging.setLoggerClass(RunwayLogger)

__version__: str = "2.7.0"
"""Version of the Python package presented as a :class:`string`.

Dynamically set upon release by `poetry-dynamic-versioning <https://github.com/mtkennerly/poetry-dynamic-versioning>`__.

"""

__version_tuple__: tuple[int, int, int] | tuple[int, int, int, str] = (2, 7, 0)
"""Version of the Python package presented as a :class:`tuple`.

Dynamically set upon release by `poetry-dynamic-versioning <https://github.com/mtkennerly/poetry-dynamic-versioning>`__.

"""



"""Runway context."""

from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any, cast

from ..compat import cached_property
from ..core.components import DeployEnvironment
from ._base import BaseContext

if TYPE_CHECKING:
    from pathlib import Path

    from .._logging import PrefixAdaptor, RunwayLogger
    from ..core.type_defs import RunwayActionTypeDef

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


def str2bool(v: str) -> bool:
    """Return boolean value of string."""
    return v.lower() in ("yes", "true", "t", "1", "on", "y")


class RunwayContext(BaseContext):
    """Runway context object."""

    command: RunwayActionTypeDef | None
    """Runway command/action being run."""

    def __init__(
        self,
        *,
        command: RunwayActionTypeDef | None = None,
        deploy_environment: DeployEnvironment | None = None,
        logger: PrefixAdaptor | RunwayLogger = LOGGER,
        work_dir: Path | None = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            command: Runway command/action being run.
            deploy_environment: The current deploy environment.
            logger: Custom logger.
            work_dir: Working directory used by Runway.

        """
        super().__init__(
            deploy_environment=deploy_environment or DeployEnvironment(),
            logger=logger,
            work_dir=work_dir,
        )
        self.command = command
        self._inject_profile_credentials()

    @cached_property
    def no_color(self) -> bool:
        """Whether to explicitly disable color output.

        Primarily applies to IaC being wrapped by Runway.

        """
        colorize = self.env.vars.get("RUNWAY_COLORIZE")  # explicitly enable/disable
        try:
            if isinstance(colorize, bool):
                # catch False
                return not colorize
            if colorize and isinstance(colorize, str):  # type: ignore
                return not str2bool(colorize)
        except ValueError:
            pass  # likely invalid RUNWAY_COLORIZE value
        return not sys.stdout.isatty()

    @cached_property
    def use_concurrent(self) -> bool:
        """Whether to use concurrent.futures or not.

        Noninteractive is required for concurrent execution to prevent weird
        user-input behavior.

        Python 3 is required because backported futures has issues with
        ProcessPoolExecutor.

        """
        if self.is_noninteractive:
            if not self.sys_info.os.is_posix:
                LOGGER.warning(
                    "parallel execution disabled; only POSIX systems are supported currently"
                )
                return False
            return True
        LOGGER.warning("parallel execution disabled; not running in CI mode")
        return False

    def copy(self) -> RunwayContext:
        """Copy the contents of this object into a new instance."""
        return self.__class__(
            command=self.command,
            deploy_environment=self.env.copy(),
            logger=self.logger,
            work_dir=self.work_dir,
        )

    def echo_detected_environment(self) -> None:
        """Print a helper note about how the environment was determined."""
        self.env.log_name()



"""AWS S3 exceptions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .....exceptions import RunwayError

if TYPE_CHECKING:
    from . import Bucket


class BucketAccessDeniedError(RunwayError):
    """Access denied to S3 Bucket."""

    bucket_name: str
    """Name of the S3 Bucket."""

    def __init__(self, bucket: Bucket) -> None:
        """Instantiate class.

        Args:
            bucket: AWS S3 Bucket object.

        """
        self.bucket_name = bucket.name
        self.message = f"access denied for bucket {bucket.name}"
        super().__init__()


class BucketNotFoundError(RunwayError):
    """S3 Bucket not found."""

    bucket_name: str
    """Name of the S3 Bucket"""

    def __init__(self, bucket: Bucket) -> None:
        """Instantiate class.

        Args:
            bucket: AWS S3 Bucket object.

        """
        self.bucket_name = bucket.name
        self.message = f"bucket {bucket.name} not found"
        super().__init__()


class S3ObjectDoesNotExistError(RunwayError):
    """Required S3 object does not exist."""

    bucket: str
    """Name of the S3 Bucket"""

    key: str
    """S3 object key."""

    uri: str
    """S3 object URI."""

    def __init__(self, bucket: str, key: str) -> None:
        """Instantiate class.

        Args:
            bucket: Name of the S3 bucket.
            key: S3 object key.

        """
        self.bucket = bucket
        self.key = key
        self.uri = f"s3://{bucket}/{key.lstrip('/')}"
        self.message = f"S3 object does not exist at path {self.uri}"
        super().__init__()



"""CFNgin blueprint base classes."""

from __future__ import annotations

import copy
import hashlib
import logging
import string
from typing import TYPE_CHECKING, Any, ClassVar

from troposphere import Output, Parameter, Ref, Template

from ...compat import cached_property
from ...mixins import DelCachedPropMixin
from ...variables import Variable
from ..exceptions import (
    InvalidUserdataPlaceholder,
    MissingVariable,
    UnresolvedBlueprintVariable,
    UnresolvedBlueprintVariables,
    ValidatorError,
    VariableTypeRequired,
)
from ..utils import read_value_from_path
from .variables.types import CFNType, TroposphereType

if TYPE_CHECKING:
    from ...context import CfnginContext
    from .type_defs import BlueprintVariableTypeDef

LOGGER = logging.getLogger(__name__)

PARAMETER_PROPERTIES = {
    "default": "Default",
    "description": "Description",
    "no_echo": "NoEcho",
    "allowed_values": "AllowedValues",
    "allowed_pattern": "AllowedPattern",
    "max_length": "MaxLength",
    "min_length": "MinLength",
    "max_value": "MaxValue",
    "min_value": "MinValue",
    "constraint_description": "ConstraintDescription",
}


class CFNParameter:
    """Wrapper around a value to indicate a CloudFormation Parameter."""

    def __init__(self, name: str, value: bool | float | list[Any] | str | Any) -> None:
        """Instantiate class.

        Args:
            name: The name of the CloudFormation Parameter.
            value: The value we're going to submit as a CloudFormation Parameter.

        """
        self.name = name
        if isinstance(value, (list, str)):
            self.value = value
        elif isinstance(value, bool):
            LOGGER.debug("converting parameter %s boolean '%s' to string", name, value)
            self.value = str(value).lower()
        elif isinstance(value, (float, int)):
            LOGGER.debug("converting parameter %s integer '%s' to string", name, value)
            self.value = str(value)
        else:
            raise TypeError(
                f"CFNParameter ({name}) value must be one of bool, float, int, str, "
                f"list[str] but got: {type(value)}"
            )

    def __repr__(self) -> str:
        """Object represented as a string."""
        return f"CFNParameter[{self.name}: {self.value}]"

    @cached_property
    def ref(self) -> Ref:
        """Ref the value of a parameter."""
        return Ref(self.name)

    def to_parameter_value(self) -> list[Any] | str:
        """Return the value to be submitted to CloudFormation."""
        return self.value


def build_parameter(name: str, properties: BlueprintVariableTypeDef) -> Parameter:
    """Build a troposphere Parameter with the given properties.

    Args:
        name: The name of the parameter.
        properties: Contains the properties that will be applied to the parameter.
            See: http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html

    Returns:
        The created parameter object.

    """
    param = Parameter(name, Type=properties.get("type"))
    for name_, attr in PARAMETER_PROPERTIES.items():
        if name_ in properties:
            setattr(param, attr, properties[name_])
    return param


def validate_variable_type(
    var_name: str,
    var_type: type[CFNType] | TroposphereType[Any] | type,
    value: Any,
) -> Any:
    """Ensure the value is the correct variable type.

    Args:
        var_name: The name of the defined variable on a blueprint.
        var_type: The type that the value should be.
        value: The object representing the value provided for the variable

    Returns:
        The appropriate value object. If the original value was of CFNType, the
        returned value will be wrapped in CFNParameter.

    Raises:
        ValueError: If the `value` isn't of `var_type` and can't be cast as
            that type, this is raised.

    """
    if isinstance(var_type, TroposphereType):
        try:
            value = var_type.create(value)
        except Exception as exc:
            raise ValidatorError(var_name, f"{var_type.resource_name}.create", value, exc) from exc
    elif issubclass(var_type, CFNType):
        value = CFNParameter(name=var_name, value=value)
    elif not isinstance(value, var_type):
        raise TypeError(
            f"Value for variable {var_name} must be of type {var_type}. Actual "
            f"type: {type(value)}"
        )
    return value


def validate_allowed_values(allowed_values: list[Any] | None, value: Any) -> bool:
    """Support a variable defining which values it allows.

    Args:
        allowed_values: A list of allowed values from the variable definition.
        value: The object representing the value provided for the variable.

    Returns:
        Boolean for whether or not the value is valid.

    """
    # ignore CFNParameter, troposphere handles these for us
    if not allowed_values or isinstance(value, CFNParameter):
        return True
    return value in allowed_values


def resolve_variable(
    var_name: str,
    var_def: BlueprintVariableTypeDef,
    provided_variable: Variable | None,
    blueprint_name: str,
) -> Any:
    """Resolve a provided variable value against the variable definition.

    Args:
        var_name: The name of the defined variable on a blueprint.
        var_def: A dictionary representing the defined variables attributes.
        provided_variable: The variable value provided to the blueprint.
        blueprint_name: The name of the blueprint that the variable is being applied to.

    Returns:
        The resolved variable value, could be any python object.

    Raises:
        MissingVariable: Raised when a variable with no default is not
            provided a value.
        UnresolvedBlueprintVariable: Raised when the provided variable is
            not already resolved.
        ValueError: Raised when the value is not the right type and cannot be
            cast as the correct type. Raised by
            :func:`runway.cfngin.blueprints.base.validate_variable_type`
        ValidatorError: Raised when a validator raises an exception. Wraps the
            original exception.

    """
    if "type" not in var_def:
        raise VariableTypeRequired(blueprint_name, var_name)
    var_type = var_def["type"]

    if provided_variable:
        if not provided_variable.resolved:
            raise UnresolvedBlueprintVariable(blueprint_name, provided_variable)

        value = provided_variable.value
    else:
        # Variable value not provided, try using the default, if it exists
        # in the definition
        if "default" not in var_def:
            raise MissingVariable(blueprint_name, var_name)
        value = var_def["default"]

    # If no validator, return the value as is, otherwise apply validator
    validator = var_def.get("validator", lambda v: v)
    try:
        value = validator(value)
    except Exception as exc:
        raise ValidatorError(var_name, validator.__name__, value, exc) from exc

    # Ensure that the resulting value is the correct type
    value = validate_variable_type(var_name, var_type, value)

    allowed_values = var_def.get("allowed_values")
    if not validate_allowed_values(allowed_values, value):
        raise ValueError(
            f"Invalid value passed to {var_name} in Blueprint {blueprint_name}. "
            f"Got '{value}', expected one of {allowed_values}."
        )

    return value


def parse_user_data(variables: dict[str, Any], raw_user_data: str, blueprint_name: str) -> str:
    """Parse the given user data and renders it as a template.

    It supports referencing template variables to create userdata
    that's supplemented with information from the stack, as commonly
    required when creating EC2 userdata files.

    Example:
        Given a raw_user_data string: ``'open file ${file}'``
        And a variables dictionary with: ``{'file': 'test.txt'}``
        parse_user_data would output: ``open file test.txt``

    Args:
        variables: Variables available to the template.
        raw_user_data: The user_data to be parsed.
        blueprint_name: The name of the blueprint.

    Returns:
        The parsed user data, with all the variables values and refs replaced
        with their resolved values.

    Raises:
        InvalidUserdataPlaceholder: Raised when a placeholder name in
            raw_user_data is not valid. E.g ``${100}`` would raise this.
        MissingVariable: Raised when a variable is in the raw_user_data that
            is not given in the blueprint

    """
    variable_values: dict[str, Any] = {}

    for key, value in variables.items():
        if isinstance(value, CFNParameter):
            variable_values[key] = value.to_parameter_value()
        else:
            variable_values[key] = value

    template = string.Template(raw_user_data)

    res = ""

    try:
        res = template.substitute(variable_values)
    except ValueError as exc:
        raise InvalidUserdataPlaceholder(blueprint_name, exc.args[0]) from exc
    except KeyError as exc:
        raise MissingVariable(blueprint_name, str(exc)) from exc

    return res


class Blueprint(DelCachedPropMixin):
    """Base implementation for rendering a troposphere template.

    Attributes:
        VARIABLES: Class variable that defines the values that can be passed
            from CFNgin to the template. These definition include metadata used
            to validate the provided value and to propagate the variable to the
            resulting CloudFormation template.
        context: CFNgin context object.
        description: The description of the CloudFormation template that will
            be generated by this Blueprint.
        mappings: CloudFormation Mappings to be added to the template during the
            rendering process.
        name: Name of the Stack that will be created by the Blueprint.
        template: Troposphere template.

    """

    VARIABLES: ClassVar[dict[str, BlueprintVariableTypeDef]] = {}

    context: CfnginContext
    description: str | None
    mappings: dict[str, dict[str, Any]] | None
    name: str
    template: Template

    def __init__(
        self,
        name: str,
        context: CfnginContext,
        *,
        description: str | None = None,
        mappings: dict[str, dict[str, Any]] | None = None,
        template: Template | None = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            name: A name for the blueprint.
            context: Context the blueprint is being executed under.
            description: The description of the CloudFormation template that will
                be generated by this blueprint.
            mappings: CloudFormation Mappings to be used in the template during
                the rendering process.
            template: Optionally, provide a preexisting Template.

        .. versionchanged:: 2.0.0
            Added :attr:`template`.

        .. versionchanged:: 2.0.0
            Class only takes 2 positional arguments.
            The rest are now keyword arguments.

        """
        self._rendered = None
        self._resolved_variables: dict[str, Any] | None = None
        self._version = None
        self.context = context
        self.description = description
        self.mappings = mappings
        self.name = name
        self.template = template or Template()

        if hasattr(self, "PARAMETERS") or hasattr(self, "LOCAL_PARAMETERS"):
            raise AttributeError(
                f"DEPRECATION WARNING: Blueprint {name} uses "
                "deprecated PARAMETERS or "
                "LOCAL_PARAMETERS, rather than VARIABLES. "
                "Please update your blueprints. See "
                "https://runway.readthedocs.io/page/cfngin/blueprints.html#variables "
                "for additional information."
            )

    @cached_property
    def cfn_parameters(self) -> dict[str, list[Any] | str]:
        """Return a dict of variables with type :class:`~runway.cfngin.blueprints.variables.types.CFNType`.

        .. versionadded:: 2.0.0

        Returns:
            Variables that need to be submitted as CloudFormation Parameters.

        """
        output: dict[str, list[Any] | str] = {}
        for key, value in self.variables.items():
            if hasattr(value, "to_parameter_value"):
                output[key] = value.to_parameter_value()
        return output

    def create_template(self) -> None:
        """Abstract method called to create a template from the blueprint."""
        raise NotImplementedError

    @property
    def defined_variables(self) -> dict[str, BlueprintVariableTypeDef]:
        """Return a copy of :attr:`VARIABLES` to avoid accidental modification of the ClassVar.

        .. versionchanged:: 2.0.0
            Changed from a method to a property.

        """
        return copy.deepcopy(self.VARIABLES)

    @property
    def output_definitions(self) -> dict[str, dict[str, Any]]:
        """Get the output definitions.

        .. versionadded:: 2.0.0

        Returns:
            Output definitions. Keys are output names, the values are dicts
            containing key/values for various output properties.

        """
        return {k: output.to_dict() for k, output in self.template.outputs.items()}

    @cached_property
    def parameter_definitions(self) -> dict[str, BlueprintVariableTypeDef]:
        """Get the parameter definitions to submit to CloudFormation.

        Any variable definition whose type is an instance of
        :class:`~runway.cfngin.blueprints.variables.types.CFNType` will
        be returned as a CloudFormation Parameter.

        .. versionadded:: 2.0.0

        Returns:
            Parameter definitions. Keys are parameter names, the values are dicts
            containing key/values for various parameter properties.

        """
        output: dict[str, BlueprintVariableTypeDef] = {}
        for var_name, attrs in self.defined_variables.items():
            var_type = attrs.get("type")
            if isinstance(var_type, type) and issubclass(var_type, CFNType):
                cfn_attrs = copy.deepcopy(attrs)
                cfn_attrs["type"] = var_type.parameter_type
                output[var_name] = cfn_attrs
        return output

    @cached_property
    def parameter_values(self) -> dict[str, list[Any] | str]:
        """Return a dict of variables with type :class:`~runway.cfngin.blueprints.variables.types.CFNType`.

        .. versionadded:: 2.0.0

        Returns:
            Variables that need to be submitted as CloudFormation Parameters.
            Will be a dictionary of <parameter name>: <parameter value>.

        """
        output: dict[str, Any] = {}
        for key, value in self.variables.items():
            try:
                output[key] = value.to_parameter_value()
            except AttributeError:
                continue
        return output

    @property
    def rendered(self) -> str:
        """Return rendered blueprint."""
        if not self._rendered:
            self._version, self._rendered = self.render_template()
        return self._rendered

    @cached_property
    def required_parameter_definitions(self) -> dict[str, BlueprintVariableTypeDef]:
        """Return all template parameters that do not have a default value.

        .. versionadded:: 2.0.0

        Returns:
            Dict of required CloudFormation Parameters for the blueprint.
            Will be a dictionary of ``<parameter name>: <parameter attributes>``.

        """
        return {
            name: attrs
            for name, attrs in self.parameter_definitions.items()
            if "Default" not in attrs
        }

    @property
    def requires_change_set(self) -> bool:
        """Return true if the underlying template has transforms."""
        return self.template.transform is not None

    @property
    def variables(self) -> dict[str, Any]:
        """Return a Dict of variables available to the Template.

        These variables will have been defined within :attr:`VARIABLES` or
        :attr:`defined_variables`.
        Any variable value that contains a Lookup will have been resolved.

        .. versionadded:: 2.0.0

        Returns:
            Variables available to the Template.

        Raises:
            UnresolvedBlueprintVariables: If variables are unresolved.

        """
        if self._resolved_variables is None:
            raise UnresolvedBlueprintVariables(self.name)
        return self._resolved_variables

    @variables.setter
    def variables(self, value: dict[str, Any]) -> None:
        """Setter for :meth:`variables`.

        .. versionadded:: 2.0.0

        """
        self._resolved_variables = value
        # clear cached properties that rely on this property
        self._del_cached_property("cfn_parameters", "parameter_values")

    @property
    def version(self) -> str:
        """Template version."""
        if not self._version:
            self._version, self._rendered = self.render_template()
        return self._version

    def add_output(self, name: str, value: Any) -> None:
        """Add an output to the template.

        Wrapper for ``self.template.add_output(Output(name, Value=value))``.

        Args:
            name: The name of the output to create.
            value: The value to put in the output.

        """
        self.template.add_output(Output(name, Value=value))

    def get_cfn_parameters(self) -> dict[str, list[Any] | str]:
        """Return a dictionary of variables with `type` :class:`CFNType`.

        .. deprecated:: 2.0.0
            Replaced by :attr:`cfn_parameters`.

        Returns:
            Variables that need to be submitted as CloudFormation Parameters.

        """
        LOGGER.warning(
            "%s.get_cfn_parameters is deprecated and will be removed in a future release",
            self.__class__.__name__,
        )
        return self.cfn_parameters

    def get_output_definitions(self) -> dict[str, dict[str, Any]]:
        """Get the output definitions.

        .. deprecated:: 2.0.0
            Replaced by :attr:`output_definitions`.

        Returns:
            Output definitions. Keys are output names, the values are dicts
            containing key/values for various output properties.

        """
        LOGGER.warning(
            "%s.get_output_definitions is deprecated and will be removed in a future release",
            self.__class__.__name__,
        )
        return self.output_definitions

    def get_parameter_definitions(self) -> dict[str, BlueprintVariableTypeDef]:
        """Get the parameter definitions to submit to CloudFormation.

        Any variable definition whose `type` is an instance of
        :class:`~runway.cfngin.blueprints.variables.types.CFNType` will
        be returned as a CloudFormation Parameter.

        .. deprecated:: 2.0.0
            Replaced by :attr:`parameter_definitions`.

        Returns:
            Parameter definitions. Keys are parameter names, the values are dicts
            containing key/values for various parameter properties.

        """
        LOGGER.warning(
            "%s.get_parameter_definitions is deprecated and will be removed in a future release",
            self.__class__.__name__,
        )
        return self.parameter_definitions

    def get_parameter_values(self) -> dict[str, list[Any] | str]:
        """Return a dict of variables with type :class:`~runway.cfngin.blueprints.variables.types.CFNType`.

        .. deprecated:: 2.0.0
            Replaced by :attr:`parameter_values`.

        Returns:
            Variables that need to be submitted as CloudFormation Parameters.
            Will be a dictionary of <parameter name>: <parameter value>.

        """
        LOGGER.warning(
            "%s.get_parameter_values is deprecated and will be removed in a future release",
            self.__class__.__name__,
        )
        return self.parameter_values

    def get_required_parameter_definitions(self) -> dict[str, BlueprintVariableTypeDef]:
        """Return all template parameters that do not have a default value.

        .. deprecated:: 2.0.0
            Replaced by :attr:`required_parameter_definitions`.

        Returns:
            Dict of required CloudFormation Parameters for the blueprint.
            Will be a dictionary of ``<parameter name>: <parameter attributes>``.

        """
        LOGGER.warning(
            "%s.get_required_parameter_definitions is deprecated and will be removed "
            "in a future release",
            self.__class__.__name__,
        )
        return self.required_parameter_definitions

    def get_variables(self) -> dict[str, Any]:
        """Return a dictionary of variables available to the template.

        These variables will have been defined within `VARIABLES` or
        `self.defined_variables`. Any variable value that contains a lookup
        will have been resolved.

        .. deprecated:: 2.0.0
            Replaced by :attr:`variables`.

        Returns:
            Variables available to the template.

        Raises:
            UnresolvedBlueprintVariables: If variables are unresolved.

        """
        LOGGER.warning(
            "%s.get_variables is deprecated and will be removed in a future release",
            self.__class__.__name__,
        )
        return self.variables

    def import_mappings(self) -> None:
        """Import mappings from CFNgin config to the blueprint."""
        if not self.mappings:
            return

        for name, mapping in self.mappings.items():
            LOGGER.debug("adding mapping %s", name)
            self.template.add_mapping(name, mapping)

    def read_user_data(self, user_data_path: str) -> str:
        """Read and parse a user_data file.

        Args:
            user_data_path: Path to the userdata file.

        """
        raw_user_data = read_value_from_path(user_data_path)
        return parse_user_data(self.variables, raw_user_data, self.name)

    def render_template(self) -> tuple[str, str]:
        """Render the Blueprint to a CloudFormation template."""
        self.import_mappings()
        self.create_template()
        if self.description:
            self.set_template_description(self.description)
        self.setup_parameters()
        rendered = self.template.to_json(indent=self.context.template_indent)
        version = hashlib.md5(rendered.encode()).hexdigest()[:8]  # noqa: S324
        return version, rendered

    def reset_template(self) -> None:
        """Reset template."""
        self.template = Template()
        self._rendered = None
        self._version = None

    def resolve_variables(self, provided_variables: list[Variable]) -> None:
        """Resolve the values of the blueprint variables.

        This will resolve the values of the `VARIABLES` with values from the
        env file, the config, and any lookups resolved.

        Args:
            provided_variables: List of provided variables.

        """
        self._resolved_variables = {}
        variable_dict = {var.name: var for var in provided_variables}
        for var_name, var_def in self.defined_variables.items():
            value = resolve_variable(var_name, var_def, variable_dict.get(var_name), self.name)
            self._resolved_variables[var_name] = value

    def set_template_description(self, description: str) -> None:
        """Add a description to the Template.

        Args:
            description: A description to be added to the resulting template.

        """
        self.template.set_description(description)

    def setup_parameters(self) -> None:
        """Add any CloudFormation parameters to the template."""
        template = self.template

        if not self.parameter_definitions:
            LOGGER.debug("no parameters defined")
            return

        for name, attrs in self.parameter_definitions.items():
            built_param = build_parameter(name, attrs)
            template.add_parameter(built_param)

    def to_json(self, variables: dict[str, Any] | None = None) -> str:
        """Render the blueprint and return the template in json form.

        Args:
            variables: Dictionary providing/overriding variable values.

        """
        variables_to_resolve: list[Variable] = []
        if variables:
            for key, value in variables.items():
                variables_to_resolve.append(Variable(key, value, "cfngin"))
        for k in self.parameter_definitions:
            if not variables or k not in variables:
                # The provided value for a CFN parameter has no effect in this
                # context (generating the CFN template), so any string can be
                # provided for its value - just needs to be something
                variables_to_resolve.append(Variable(k, "unused_value", "cfngin"))  # noqa: PERF401
        self.resolve_variables(variables_to_resolve)

        return self.render_template()[1]



"""Runway Static Site Module parameters."""

from ._models import (
    RunwayStaticSiteCustomErrorResponseDataModel,
    RunwayStaticSiteLambdaFunctionAssociationDataModel,
    RunwayStaticSiteModuleParametersDataModel,
)

__all__ = [
    "RunwayStaticSiteCustomErrorResponseDataModel",
    "RunwayStaticSiteLambdaFunctionAssociationDataModel",
    "RunwayStaticSiteModuleParametersDataModel",
]



"""AWS Lambda hook."""

from __future__ import annotations

import hashlib
import json
import locale
import logging
import os
import shutil
import stat
import subprocess
import sys
import tempfile
from io import BytesIO as StringIO
from pathlib import Path
from shutil import copyfile
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)
from zipfile import ZIP_DEFLATED, ZipFile

import botocore
import botocore.exceptions
import docker
import docker.types
import formic
from troposphere.awslambda import Code
from typing_extensions import Literal, TypedDict

from ..exceptions import InvalidDockerizePipConfiguration, PipenvError, PipError
from ..utils import ensure_s3_bucket

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from docker.models.images import Image
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.literals import ObjectCannedACLType
    from mypy_boto3_s3.type_defs import HeadObjectOutputTypeDef
    from typing_extensions import TypeAlias

    from ...context import CfnginContext
    from ..providers.aws.default import Provider

# mask to retrieve only UNIX file permissions from the external attributes
# field of a ZIP entry.
ZIP_PERMS_MASK = (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) << 16

LOGGER = logging.getLogger(__name__)

# list from python tags of https://hub.docker.com/r/lambci/lambda/tags
SUPPORTED_RUNTIMES = ["python3.7", "python3.8"]

DockerizePipArgTypeDef: TypeAlias = (
    'bool | Literal["false", "False", "no", "No", "non-linux", "true", "True", "yes", "Yes"] | None'
)


def copydir(
    source: str,
    destination: str,
    includes: list[str],
    excludes: list[str] | None = None,
    follow_symlinks: bool = False,
) -> None:
    """Extend the functionality of shutil.

    Correctly copies files and directories in a source directory.

    Args:
        source: Source directory.
        destination: Destination directory.
        includes: Glob patterns for files to include.
        excludes: Glob patterns for files to exclude.
        follow_symlinks: If true, symlinks will be included in the resulting zip
            file.

    """
    files = _find_files(source, includes, excludes, follow_symlinks)

    def _mkdir(dir_name: str) -> None:
        """Recursively create directories."""
        parent = os.path.dirname(dir_name)  # noqa: PTH120
        if not os.path.isdir(parent):  # noqa: PTH112
            _mkdir(parent)
        LOGGER.debug("creating directory: %s", dir_name)
        os.mkdir(dir_name)  # noqa: PTH102

    for file_name in files:
        src = os.path.join(source, file_name)  # noqa: PTH118
        dest = os.path.join(destination, file_name)  # noqa: PTH118
        try:
            LOGGER.debug('copying file "%s" to "%s"', src, dest)
            copyfile(src, dest)
        except OSError:
            _mkdir(os.path.dirname(dest))  # noqa: PTH120
            copyfile(src, dest)


def find_requirements(root: str) -> dict[str, bool] | None:
    """Identify Python requirement files.

    Args:
        root: Path that should be searched for files.

    Returns:
        Name of supported requirements file and whether it was found.
        If none are found, ``None`` is returned.

    """
    findings = {
        file_name: os.path.isfile(os.path.join(root, file_name))  # noqa: PTH118, PTH113
        for file_name in ["requirements.txt", "Pipfile", "Pipfile.lock"]
    }

    if not sum(findings.values()):
        return None
    return findings


def should_use_docker(dockerize_pip: DockerizePipArgTypeDef = None) -> bool:
    """Assess if Docker should be used based on the value of args.

    Args:
        dockerize_pip: Value to assess if Docker should be used for pip.

    """
    if isinstance(dockerize_pip, bool):
        return dockerize_pip

    if isinstance(dockerize_pip, str):
        if dockerize_pip == "non-linux" and not sys.platform.startswith("linux"):
            return True
        try:
            return bool(str2bool(dockerize_pip))
        except ValueError:
            pass
    return False


def str2bool(v: str) -> bool:
    """Return boolean value of string."""
    return v.lower() in ("yes", "true", "t", "1", "on", "y")


def _zip_files(files: Iterable[str], root: str) -> tuple[bytes, str]:
    """Generate a ZIP file in-memory from a list of files.

    Files will be stored in the archive with relative names, and have their
    UNIX permissions forced to 755 or 644 (depending on whether they are
    user-executable in the source filesystem).

    Args:
        files: file names to add to the archive, relative to ``root``.
        root: base directory to retrieve files from.

    Returns:
        Content of the ZIP file as a byte string and calculated hash of all the files.

    """
    zip_data = StringIO()
    files = list(files)  # create copy of list also converts generator to list
    with ZipFile(zip_data, "w", ZIP_DEFLATED) as zip_file:
        for file_name in files:
            zip_file.write(os.path.join(root, file_name), file_name)  # noqa: PTH118

        # Fix file permissions to avoid any issues - only care whether a file
        # is executable or not, choosing between modes 755 and 644 accordingly.
        for zip_entry in zip_file.filelist:
            perms = (zip_entry.external_attr & ZIP_PERMS_MASK) >> 16
            new_perms = 0o755 if perms & stat.S_IXUSR != 0 else 0o644
            if new_perms != perms:
                LOGGER.debug("fixing perms: %s: %o => %o", zip_entry.filename, perms, new_perms)
                new_attr = (zip_entry.external_attr & ~ZIP_PERMS_MASK) | (new_perms << 16)
                zip_entry.external_attr = new_attr

    contents = zip_data.getvalue()
    zip_data.close()
    content_hash = _calculate_hash(files, root)

    return contents, content_hash


def _calculate_hash(files: Iterable[str], root: str) -> str:
    """Return a hash of all of the given files at the given root.

    Args:
        files: file names to include in the hash calculation,
            relative to ``root``.
        root: base directory to analyze files in.

    """
    file_hash = hashlib.md5()  # noqa: S324
    for file_name in sorted(files):
        file_path = os.path.join(root, file_name)  # noqa: PTH118
        file_hash.update((file_name + "\0").encode())
        with open(file_path, "rb") as file_:  # noqa: PTH123
            for chunk in iter(lambda: file_.read(4096), ""):
                if not chunk:
                    break
                file_hash.update(chunk)
            file_hash.update(b"\0")

    return file_hash.hexdigest()


def _find_files(
    root: str,
    includes: list[str] | str,
    excludes: list[str] | None = None,
    follow_symlinks: bool = False,
) -> Iterator[str]:
    """List files inside a directory based on include and exclude rules.

    This is a more advanced version of `glob.glob`, that accepts multiple
    complex patterns.

    Args:
        root: base directory to list files from.
        includes: inclusion patterns. Only files matching those patterns will be
            included in the result.
        excludes: exclusion patterns. Files matching those patterns will be
            excluded from the result. Exclusions take precedence over inclusions.
        follow_symlinks: If true, symlinks will be included in the resulting zip file.

    Yields:
        File names relative to the root.

    Note:
        Documentation for the patterns can be found at
        http://www.aviser.asia/formic/doc/index.html

    """
    root = os.path.abspath(root)  # noqa: PTH100
    file_set = formic.FileSet(
        directory=root, include=includes, exclude=excludes, symlinks=follow_symlinks
    )
    yield from file_set.qualified_files(absolute=False)


def _zip_from_file_patterns(
    root: str, includes: list[str], excludes: list[str], follow_symlinks: bool
) -> tuple[bytes, str]:
    """Generate a ZIP file in-memory from file search patterns.

    Args:
        root: Base directory to list files from.
        includes: Inclusion patterns. Only files  matching those patterns will be
            included in the result.
        excludes: Exclusion patterns. Files matching those patterns will be
            excluded from the result. Exclusions take precedence over inclusions.
        follow_symlinks: If true, symlinks will be included in the resulting zip file.

    See Also:
        :func:`_zip_files`, :func:`_find_files`.

    Raises:
        RuntimeError: when the generated archive would be empty.

    """
    LOGGER.info("base directory: %s", root)

    files = list(_find_files(root, includes, excludes, follow_symlinks))
    if not files:
        raise RuntimeError(
            "Empty list of files for Lambda payload. Check "
            "your include/exclude options for errors."
        )

    LOGGER.info("adding %d files:", len(files))

    for file_name in files:
        LOGGER.debug(" + %s", file_name)

    return _zip_files(files, root)


def handle_requirements(
    package_root: str,
    dest_path: str,
    requirements: dict[str, bool],
    pipenv_timeout: int = 300,
    python_path: str | None = None,
    use_pipenv: bool = False,
) -> str:
    """Use the correct requirements file.

    Args:
        package_root: Base directory containing a requirements file.
        dest_path: Where to output the requirements file if one needs to be created.
        requirements: Map of requirement file names and whether they exist.
        pipenv_timeout: Seconds to wait for a subprocess to complete.
        python_path: Explicit python interpreter to be used. Requirement file
            generators must be installed and executable using ``-m`` if provided.
        use_pipenv: Explicitly use pipenv to export a Pipfile as requirements.txt.

    Returns:
        Path to the final requirements.txt

    Raises:
        NotImplementedError: When a requirements file is not found. This
            should never be encountered but is included just in case.

    """
    if use_pipenv:
        LOGGER.info("explicitly using pipenv")
        return _handle_use_pipenv(
            package_root=package_root,
            dest_path=dest_path,
            python_path=python_path,
            timeout=pipenv_timeout,
        )
    if requirements["requirements.txt"]:
        LOGGER.info("using requirements.txt for dependencies")
        return os.path.join(dest_path, "requirements.txt")  # noqa: PTH118
    if requirements["Pipfile"] or requirements["Pipfile.lock"]:
        LOGGER.info("using pipenv for dependencies")
        return _handle_use_pipenv(
            package_root=package_root,
            dest_path=dest_path,
            python_path=python_path,
            timeout=pipenv_timeout,
        )
    # This point should never be reached under normal operation since a
    # requirements file of some sort must have been found in another step
    # of the process but just in case it does happen, raise an error.
    raise NotImplementedError("Unable to handle missing requirements file.")


def _handle_use_pipenv(
    package_root: str,
    dest_path: str,
    python_path: str | None = None,
    timeout: int = 300,
) -> str:
    """Create requirements file from Pipfile.

    Args:
        package_root: Base directory to generate requirements from.
        dest_path: Where to output the requirements file.
        python_path: Explicit python interpreter to be used. pipenv must be
            installed and executable using ``-m`` if provided.
        timeout: Seconds to wait for process to complete.

    Raises:
        PipenvError: Non-zero exit code returned by pipenv process.

    """
    if getattr(sys, "frozen", False):
        LOGGER.error("pipenv can only be used with python installed from PyPi")
        sys.exit(1)
    LOGGER.info("creating requirements.txt from Pipfile...")
    req_path = os.path.join(dest_path, "requirements.txt")  # noqa: PTH118
    cmd = ["pipenv", "lock", "--requirements", "--keep-outdated"]
    if python_path:
        cmd.insert(0, python_path)
        cmd.insert(1, "-m")
    with (
        open(req_path, "w", encoding="utf-8") as requirements,  # noqa: PTH123
        subprocess.Popen(
            cmd, cwd=package_root, stdout=requirements, stderr=subprocess.PIPE
        ) as pipenv_process,
    ):
        _stdout, stderr = pipenv_process.communicate(timeout=timeout)
        if pipenv_process.returncode == 0:
            return req_path
        stderr = stderr.decode("UTF-8")
        LOGGER.error('"%s" failed with the following output:\n%s', " ".join(cmd), stderr)
        raise PipenvError


def dockerized_pip(  # noqa: C901, PLR0912
    work_dir: str,
    client: docker.DockerClient | None = None,
    runtime: str | None = None,
    docker_file: str | None = None,
    docker_image: str | None = None,
    python_dontwritebytecode: bool = False,
    **_: Any,
) -> None:
    """Run pip with docker.

    Args:
        work_dir: Work directory for docker.
        client: Custom docker client.
        runtime: Lambda runtime. Must provide one of ``runtime``,
            ``docker_file``, or ``docker_image``.
        docker_file: Path to a Dockerfile to build an image.
            Must provide one of ``runtime``, ``docker_file``, or ``docker_image``.
        docker_image: Local or remote docker image to use.
            Must provide one of ``runtime``, ``docker_file``, or ``docker_image``.
        python_dontwritebytecode: Don't write bytecode.

    """
    # TODO (craig): use kwargs to pass args to docker for advanced config
    if bool(docker_file) + bool(docker_image) + bool(runtime) != 1:
        # exactly one of these is needed. converting to bool will give us a
        # 'False' (0) for 'None' and 'True' (1) for anything else.
        raise InvalidDockerizePipConfiguration(
            "exactly only one of [docker_file, docker_file, runtime] must be provided"
        )

    if not client:
        client = docker.from_env()

    if docker_file:
        if not os.path.isfile(docker_file):  # noqa: PTH113
            raise ValueError(f'could not find docker_file "{docker_file}"')
        LOGGER.info('building docker image from "%s"', docker_file)
        response = cast(
            "Image | tuple[Image, Iterator[dict[str, str]]]",
            client.images.build(
                path=os.path.dirname(docker_file),  # noqa: PTH120
                dockerfile=os.path.basename(docker_file),  # noqa: PTH119
                forcerm=True,
            ),
        )
        # the response can be either a tuple of (Image, Generator[dict[str, str]])
        # or just Image depending on API version.
        if isinstance(response, tuple):
            docker_image = response[0].id
            for log_msg in response[1]:
                if log_msg.get("stream"):
                    LOGGER.info(log_msg["stream"].strip("\n"))
        else:
            docker_image = response.id
        LOGGER.info('docker image "%s" created', docker_image)
    if runtime:
        if runtime not in SUPPORTED_RUNTIMES:
            raise ValueError(f'invalid runtime "{runtime}" must be one of {SUPPORTED_RUNTIMES}')
        docker_image = f"lambci/lambda:build-{runtime}"
        LOGGER.debug('selected docker image "%s" based on provided runtime', docker_image)

    if sys.platform.lower() == "win32":
        LOGGER.debug("formatted docker mount path for Windows")
        work_dir = work_dir.replace("\\", "/")

    work_dir_mount = docker.types.Mount(target="/var/task", source=work_dir, type="bind")
    pip_cmd = "python -m pip install -t /var/task -r /var/task/requirements.txt"

    LOGGER.info('using docker image "%s" to build deployment package...', docker_image)

    docker_run_args: dict[str, Any] = {}
    if python_dontwritebytecode:
        docker_run_args["environment"] = "1"

    container = client.containers.run(
        image=cast(str, docker_image),
        command=["/bin/sh", "-c", pip_cmd],
        auto_remove=True,
        detach=True,
        mounts=[work_dir_mount],
        **docker_run_args,
    )

    # 'stream' creates a blocking generator that allows for real-time logs.
    # this loop ends when the container 'auto_remove's itself.
    for log in container.logs(stdout=True, stderr=True, stream=True, tail=0):
        # without strip there are a bunch blank lines in the output
        LOGGER.info(log.decode().strip())


def _pip_has_no_color_option(python_path: str) -> bool:
    """Return boolean on whether pip is new enough to have --no-color option.

    pip v10 introduced this option, and it's used to minimize the effect of
    pip falsely indicating errors like the following:

    "ERROR: awscli 1.18.64 has requirement rsa<=3.5.0,>=3.1.2, but you'll
    have rsa 4.0 which is incompatible."

    An error like that (colored in red) will appear when there's a conflict
    between the host & target environments, which is not helpful in the
    context of building a Lambda zip that will never interact with the
    running python system.

    Ideally this mitigation (and this function by association) can be removed
    by an enhancement or replacement of pip for building the packages.

    """
    try:
        pip_version_string = subprocess.check_output(
            [
                python_path,
                "-c",
                "from __future__ import print_function;import pip;print(pip.__version__)",
            ]
        )
        if isinstance(pip_version_string, bytes):  # type: ignore
            pip_version_string = pip_version_string.decode()
        if int(pip_version_string.split(".", maxsplit=1)[0]) > 10:
            return True
    except (AttributeError, ValueError, subprocess.CalledProcessError):
        LOGGER.debug("error checking pip version; assuming it to be pre-v10")
    return False


# TODO (kyle): refactor logic to breakup logic into smaller chunks
def _zip_package(  # noqa: PLR0915, PLR0912, C901, D417
    package_root: str,
    *,
    dockerize_pip: DockerizePipArgTypeDef = False,
    excludes: list[str] | None = None,
    follow_symlinks: bool = False,
    includes: list[str],
    pipenv_timeout: int = 300,
    python_dontwritebytecode: bool = False,
    python_exclude_bin_dir: bool = False,
    python_exclude_setuptools_dirs: bool = False,
    python_path: str | None = None,
    requirements_files: dict[str, bool],
    use_pipenv: bool = False,
    work_dir: Path,
    **kwargs: Any,
) -> tuple[bytes, str]:
    """Create zip file in memory with package dependencies.

    Args:
        package_root: Base directory to copy files from.
        dockerize_pip: Whether to use docker or under what conditions docker will
            be used to run ``pip``.
        excludes: Exclusion patterns. Files matching those patterns will be
            excluded from the result. Exclusions take precedence over inclusions.
        follow_symlinks: If true, symlinks will be included in the resulting zip file.
        includes: Inclusion patterns. Only files  matching those patterns will be
            included in the result.
        pipenv_timeout: pipenv timeout in seconds.
        python_dontwritebytecode: Done write byte code.
        python_exclude_bin_dir: Exclude bin directory.
        python_exclude_setuptools_dirs: Exclude setuptools directories.
        python_path: Explicit python interpreter to be used. pipenv must be
            installed and executable using ``-m`` if provided.
        requirements_files: Map of requirement file names and whether they exist.
        use_pipenv: Whether to use pipenv to export a Pipfile as requirements.txt.
        work_dir: Working directory.

    Returns:
        Content of the ZIP file as a byte string and calculated hash of all the files

    """
    kwargs.setdefault("pipenv_timeout", 300)

    if not work_dir.is_dir():
        work_dir.mkdir(parents=True)

    # exclude potential virtual environments in the package
    excludes = excludes or []
    excludes.append(".venv/")

    tmpdir = tempfile.TemporaryDirectory(prefix="cfngin", dir=work_dir)
    tmp_req = os.path.join(tmpdir.name, "requirements.txt")  # noqa: PTH118
    copydir(package_root, tmpdir.name, includes, excludes, follow_symlinks)
    tmp_req = handle_requirements(
        package_root=package_root,
        dest_path=tmpdir.name,
        requirements=requirements_files,
        python_path=python_path,
        use_pipenv=use_pipenv,
        pipenv_timeout=pipenv_timeout,
    )

    if should_use_docker(dockerize_pip):
        dockerized_pip(tmpdir.name, **kwargs)
    else:
        tmp_script = Path(tmpdir.name) / "__runway_run_pip_install.py"
        pip_cmd = [
            python_path or sys.executable,
            "-m",
            "pip",
            "install",
            "--target",
            tmpdir.name,
            "--requirement",
            tmp_req,
            "--no-color",
        ]

        subprocess_args: dict[str, Any] = {}
        if python_dontwritebytecode:
            subprocess_args["env"] = dict(os.environ, PYTHONDONTWRITEBYTECODE="1")

        # Pyinstaller build or explicit python path
        if getattr(sys, "frozen", False) and not python_path:
            script_contents = os.linesep.join(
                [
                    "import runpy",
                    "from runway.utils imports argv",
                    f"with argv(*{json.dumps(pip_cmd[2:])}):",
                    '   runpy.run_module("pip", run_name="__main__")\n',
                ]
            )
            tmp_script.write_text(
                script_contents,
                encoding=locale.getpreferredencoding(do_setlocale=False),
            )
            cmd = [sys.executable, "run-python", str(tmp_script)]
        else:
            if not _pip_has_no_color_option(pip_cmd[0]):
                pip_cmd.remove("--no-color")
            cmd = pip_cmd

        LOGGER.info(
            "The following output from pip may include incompatibility errors. "
            "These can generally be ignored (pip will erroneously warn "
            "about conflicts between the packages in your Lambda zip and "
            "your host system)."
        )

        try:
            subprocess.check_call(cmd, **subprocess_args)
        except subprocess.CalledProcessError:
            raise PipError from None
        finally:
            if tmp_script.is_file():
                tmp_script.unlink()

    if python_exclude_bin_dir and os.path.isdir(  # noqa: PTH112
        os.path.join(tmpdir.name, "bin")  # noqa: PTH118
    ):
        LOGGER.debug("Removing python /bin directory from Lambda files")
        shutil.rmtree(os.path.join(tmpdir.name, "bin"))  # noqa: PTH118
    if python_exclude_setuptools_dirs:
        for i in os.listdir(tmpdir.name):
            if i.endswith((".egg-info", ".dist-info")):
                LOGGER.debug("Removing directory %s from Lambda files", i)
                shutil.rmtree(os.path.join(tmpdir.name, i))  # noqa: PTH118

    req_files = _find_files(tmpdir.name, includes="**", follow_symlinks=False)
    contents, content_hash = _zip_files(req_files, tmpdir.name)
    if sys.version_info.major < 3:
        remove_error = OSError
    else:
        remove_error = PermissionError
    try:
        tmpdir.cleanup()
    except remove_error:
        LOGGER.warning(
            'Error removing temporary Lambda build directory "%s", '
            "likely due to root-owned files it in. Delete it manually to "
            "reclaim space",
            tmpdir.name,
        )
    return contents, content_hash


def _head_object(s3_conn: S3Client, bucket: str, key: str) -> HeadObjectOutputTypeDef | None:
    """Retrieve information about an object in S3 if it exists.

    Args:
        s3_conn: S3 connection to use for operations.
        bucket: name of the bucket containing the key.
        key: name of the key to lookup.

    Returns:
        S3 object information, or None if the object does not exist.
        See the AWS documentation for explanation of the contents.

    Raises:
        botocore.exceptions.ClientError: any error from boto3 other than key
            not found is passed through.

    """
    try:
        return s3_conn.head_object(Bucket=bucket, Key=key)
    except botocore.exceptions.ClientError as err:
        if err.response["Error"]["Code"] == "404":
            return None
        raise


def _upload_code(
    s3_conn: S3Client,
    bucket: str,
    prefix: str,
    name: str,
    contents: bytes | str,
    content_hash: str,
    payload_acl: ObjectCannedACLType,
) -> Code:
    """Upload a ZIP file to S3 for use by Lambda.

    The key used for the upload will be unique based on the checksum of the
    contents. No changes will be made if the contents in S3 already match the
    expected contents.

    Args:
        s3_conn: S3 connection to use for operations.
        bucket: name of the bucket to create.
        prefix: S3 prefix to prepend to the constructed key name for
            the uploaded file
        name: desired name of the Lambda function. Will be used to construct a
            key name for the uploaded file.
        contents: byte string with the content of the file upload.
        content_hash: md5 hash of the contents to be uploaded.
        payload_acl: The canned S3 object ACL to be applied to the uploaded payload.

    Returns:
        CloudFormation Lambda Code object, pointing to the uploaded payload in S3.

    Raises:
        botocore.exceptions.ClientError: any error from boto3 is passed
            through.

    """
    LOGGER.debug("ZIP hash: %s", content_hash)
    key = f"{prefix}lambda-{name}-{content_hash}.zip"

    if _head_object(s3_conn, bucket, key):
        LOGGER.info("object already exists; not uploading: %s", key)
    else:
        LOGGER.info("uploading object: %s", key)
        s3_conn.put_object(
            Bucket=bucket,
            Key=key,
            Body=contents.encode() if isinstance(contents, str) else contents,
            ContentType="application/zip",
            ACL=payload_acl,
        )

    return Code(S3Bucket=bucket, S3Key=key)


def _check_pattern_list(
    patterns: list[str] | str | None,
    key: str,
    default: list[str] | None = None,
) -> list[str] | None:
    """Validate file search patterns from user configuration.

    Acceptable input is a string (which will be converted to a singleton list),
    a list of strings, or anything falsy (such as None or an empty dictionary).
    Empty or unset input will be converted to a default.

    Args:
        patterns: Input from user configuration (YAML).
        key (str): Name of the configuration key the input came from,
            used for error display purposes.
        default: Value to return in case the input is empty or unset.

    Returns:
        Validated list of patterns.

    Raises:
        ValueError: If the input is unacceptable.

    """
    if not patterns:
        return default

    if isinstance(patterns, str):
        return [patterns]

    if isinstance(patterns, list) and all(isinstance(p, str) for p in patterns):  # type: ignore
        return patterns

    raise ValueError(f"Invalid file patterns in key '{key}': must be a string or list of strings")


class _UploadFunctionOptionsTypeDef(TypedDict):
    """Type definition for the "options" argument of _upload_function.

    Attributes:
        include: File patterns to include in the payload.
        exclude: File patterns to exclude from the payload.
        path: Base path to retrieve files from. If not absolute, it
            will be interpreted as relative to the CFNgin configuration file
            directory, then converted to an absolutepath.

    """

    exclude: list[str] | None
    include: list[str] | None
    path: str


def _upload_function(
    s3_conn: S3Client,
    bucket: str,
    prefix: str,
    name: str,
    options: _UploadFunctionOptionsTypeDef,
    follow_symlinks: bool,
    payload_acl: ObjectCannedACLType,
    sys_path: str,
    work_dir: Path,
) -> Code:
    """Build a Lambda payload from user configuration and uploads it to S3.

    Args:
        s3_conn: S3 connection to use for operations.
        bucket: name of the bucket to upload to.
        prefix: S3 prefix to prepend to the constructed key name for
            the uploaded file
        name: Desired name of the Lambda function. Will be used to
            construct a key name for the uploaded file.
        options: Configuration for how to build the payload.
        follow_symlinks: If true, symlinks will be included in the
            resulting zip file
        payload_acl: The canned S3 object ACL to be applied to the
            uploaded payload
        sys_path: Path that all actions are relative to.
        work_dir: Working directory used by the hook.

    Returns:
        CloudFormation AWS Lambda Code object, pointing to the uploaded object in S3.

    Raises:
        ValueError: If any configuration is invalid.
        botocore.exceptions.ClientError: Any error from boto3 is passed
            through.

    """
    try:
        root = os.path.expanduser(options["path"])  # noqa: PTH111
    except KeyError as exc:
        raise ValueError(f"missing required property '{exc.args[0]}' in function '{name}'") from exc

    includes = _check_pattern_list(options.get("include"), "include", default=["**"])
    excludes = _check_pattern_list(options.get("exclude"), "exclude", default=[])

    LOGGER.debug("processing function: %s", name)

    # os.path.join will ignore other parameters if the right-most one is an
    # absolute path, which is exactly what we want.
    if not os.path.isabs(root):  # noqa: PTH117
        root = os.path.abspath(os.path.join(sys_path, root))  # noqa: PTH118, PTH100
    requirements_files = find_requirements(root)
    if requirements_files:
        zip_contents, content_hash = _zip_package(
            root,
            includes=cast("list[str]", includes),
            excludes=excludes,
            follow_symlinks=follow_symlinks,
            requirements_files=requirements_files,
            work_dir=work_dir,
            **options,
        )
    else:
        zip_contents, content_hash = _zip_from_file_patterns(
            root, cast("list[str]", includes), cast("list[str]", excludes), follow_symlinks
        )

    return _upload_code(s3_conn, bucket, prefix, name, zip_contents, content_hash, payload_acl)


def select_bucket_region(
    custom_bucket: str | None,
    hook_region: str | None,
    cfngin_bucket_region: str | None,
    provider_region: str,
) -> str:
    """Return the appropriate region to use when uploading functions.

    Select the appropriate region for the bucket where lambdas are uploaded in.

    Args:
        custom_bucket: The custom bucket name provided by the `bucket` kwarg of
            the aws_lambda hook, if provided.
        hook_region: The contents of the `bucket_region` argument to the hook.
        cfngin_bucket_region: The contents of the ``cfngin_bucket_region`` global
            setting.
        provider_region: The region being used by the provider.

    Returns:
        The appropriate region string.

    """
    region = None
    region = hook_region if custom_bucket else cfngin_bucket_region
    return region or provider_region


def upload_lambda_functions(  # noqa: D417
    context: CfnginContext, provider: Provider, **kwargs: Any
) -> dict[str, Any]:
    """Build Lambda payloads from user configuration and upload them to S3.

    Constructs ZIP archives containing files matching specified patterns for
    each function, uploads the result to Amazon S3, then stores objects (of
    type :class:`troposphere.awslambda.Code`) in the context's hook data,
    ready to be referenced in blueprints.

    Configuration consists of some global options, and a dictionary of function
    specifications. In the specifications, each key indicating the name of the
    function (used for generating names for artifacts), and the value
    determines what files to include in the ZIP (see more details below).

    Payloads are uploaded to either a custom bucket or the CFNgin default
    bucket, with the key containing it's checksum, to allow repeated uploads
    to be skipped in subsequent runs.

    The configuration settings are documented as keyword arguments below.

    Args:
        provider: Provider instance. (passed in by CFNgin)
        context: Context instance. (passed in by CFNgin)

    Keyword Args:
        bucket (str | None): Custom bucket to upload functions to.
            Omitting it will cause the default CFNgin bucket to be used.
        bucket_region (str | None): The region in which the bucket should
            exist. If not given, the region will be either be that of the
            global ``cfngin_bucket_region`` setting, or else the region in
            use by the provider.
        prefix (str | None): S3 key prefix to prepend to the uploaded
            zip name.
        follow_symlinks (str | None): Will determine if symlinks should
            be followed and included with the zip artifact. (*default:*
            ``False``)
        payload_acl (str | None): The canned S3 object ACL to be applied
            to the uploaded payload. (*default: private*)
        functions (dict[str, Any]): Configurations of desired payloads to
            build. Keys correspond to function names, used to derive key
            names for the payload. Each value should itself be a dictionary,
            with the following data:

            **docker_file (str | None)**
                Path to a local DockerFile that will be built and used for
                ``dockerize_pip``. Must provide exactly one of ``docker_file``,
                ``docker_image``, or ``runtime``.

            **docker_image (str | None)**
                Custom Docker image to use  with ``dockerize_pip``. Must
                provide exactly one of ``docker_file``, ``docker_image``, or
                ``runtime``.

            **dockerize_pip (bool | str | None)**
                Whether to use Docker when preparing package dependencies with
                pip. Can be set to True/False or the special string 'non-linux'
                which will only run on non Linux systems. To use this option
                Docker must be installed.

            **exclude (str | list[str] | None)**
                Pattern or list of patterns of files to exclude from the
                payload. If provided, any files that match will be ignored,
                regardless of whether they match an inclusion pattern.

                Commonly ignored files are already excluded by default,
                such as ``.git``, ``.svn``, ``__pycache__``, ``*.pyc``,
                ``.gitignore``, etc.

            **include (str | list[str] | None)**
                Pattern or list of patterns of files to include in the
                payload. If provided, only files that match these
                patterns will be included in the payload.

                Omitting it is equivalent to accepting all files that are
                not otherwise excluded.

            **path (str)**
                Base directory of the Lambda function payload content.
                If it not an absolute path, it will be considered relative
                to the directory containing the CFNgin configuration file
                in use.

                Files in this directory will be added to the payload ZIP,
                according to the include and exclude patterns. If no
                patterns are provided, all files in this directory
                (respecting default exclusions) will be used.

                Files are stored in the archive with path names relative to
                this directory. So, for example, all the files contained
                directly under this directory will be added to the root of
                the ZIP file.

            **pipenv_lock_timeout (int | None)**
                Time in seconds to wait while creating lock file with pipenv.

            **pipenv_timeout (int | None)**
                Time in seconds to wait while running pipenv.

            **python_path (str | None)**
                Absolute path to a python interpreter to use for
                ``pip``/``pipenv`` actions. If not provided, the current
                python interpreter will be used for ``pip`` and ``pipenv``
                will be used from the current ``$PATH``.

            **runtime (str | None)**
                Runtime of the AWS Lambda Function being uploaded. Used with
                ``dockerize_pip`` to automatically select the appropriate
                Docker image to use. Must provide exactly one of
                ``docker_file``, ``docker_image``, or ``runtime``.

            **use_pipenv (bool)**
                Explicitly use Pipfile/Pipfile.lock to prepare package
                dependencies even if a requirements.txt file is found.

    Examples:
        .. Hook configuration.
        .. code-block:: yaml

            pre_deploy:
              - path: runway.cfngin.hooks.aws_lambda.upload_lambda_functions
                required: true
                enabled: true
                data_key: lambda
                args:
                  bucket: custom-bucket
                  follow_symlinks: true
                  prefix: cloudformation-custom-resources/
                  payload_acl: authenticated-read
                  functions:
                    MyFunction:
                      path: ./lambda_functions
                      dockerize_pip: non-linux
                      use_pipenv: true
                      runtime: python3.9
                      include:
                        - '*.py'
                        - '*.txt'
                      exclude:
                        - '*.pyc'
                        - test/

        .. Blueprint usage
        .. code-block:: python

            from troposphere.awslambda import Function
            from runway.cfngin.blueprints.base import Blueprint

            class LambdaBlueprint(Blueprint):
                def create_template(self):
                    code = self.context.hook_data['lambda']['MyFunction']

                    self.template.add_resource(
                        Function(
                            'MyFunction',
                            Code=code,
                            Handler='my_function.handler',
                            Role='...',
                            Runtime='python2.7'
                        )
                    )

    """
    LOGGER.warning(
        "%s is deprecated and will be removed in a future release - "
        "see documentation for replacement",
        __name__,
    )
    # TODO (craig): add better handling for misconfiguration (e.g. forgetting function names)
    # TODO (craig): support defining dockerize_pip options at the top level of args
    custom_bucket = cast(str, kwargs.get("bucket", ""))
    if not custom_bucket:
        if not context.bucket_name:
            raise ValueError("hook requires bucket argument or top-level cfngin_hook")
        bucket_name = context.bucket_name
        LOGGER.info("using default bucket from CFNgin: %s", bucket_name)
    else:
        bucket_name = custom_bucket
        LOGGER.info("using custom bucket: %s", bucket_name)

    custom_bucket_region = cast(str, kwargs.get("bucket_region", ""))
    if not custom_bucket and custom_bucket_region:
        raise ValueError("Cannot specify `bucket_region` without specifying `bucket`.")

    bucket_region = select_bucket_region(
        custom_bucket,
        custom_bucket_region,
        context.config.cfngin_bucket_region,
        provider.region or "us-east-1",
    )

    # Check if we should walk / follow symlinks
    follow_symlinks = kwargs.get("follow_symlinks", False)
    if not isinstance(follow_symlinks, bool):
        raise ValueError("follow_symlinks option must be a boolean")

    # Check for S3 object acl. Valid values from:
    # https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
    payload_acl = cast("ObjectCannedACLType", kwargs.get("payload_acl", "private"))

    # Always use the global client for s3
    session = context.get_session(region=bucket_region)
    s3_client = session.client("s3")

    ensure_s3_bucket(s3_client, bucket_name, bucket_region)

    prefix = kwargs.get("prefix", "")

    results: dict[str, Any] = {}
    for name, options in kwargs["functions"].items():
        sys_path = (
            os.path.dirname(context.config_path)  # noqa: PTH120
            if os.path.isfile(context.config_path)  # noqa: PTH113
            else context.config_path
        )
        results[name] = _upload_function(
            s3_client,
            bucket_name,
            prefix,
            name,
            options,
            follow_symlinks,
            payload_acl,
            str(sys_path),
            work_dir=context.work_dir,
        )

    return results



"""User Pool Client Updater.

Responsible for updating the User Pool Client with the generated
distribution url + callback url paths.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from ...base import HookArgsBaseModel

if TYPE_CHECKING:
    from .....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    alternate_domains: list[str]
    """A list of any alternate domains that need to be listed with the primary
    distribution domain.

    """

    client_id: str
    """Client ID."""

    distribution_domain: str
    """Distribution domain."""

    oauth_scopes: list[str]
    """A list of all available validation scopes for oauth."""

    redirect_path_sign_in: str
    """The redirect path after sign in."""

    redirect_path_sign_out: str
    """The redirect path after sign out."""

    supported_identity_providers: list[str] = []
    """Supported identity providers."""


def get_redirect_uris(
    domains: list[str], redirect_path_sign_in: str, redirect_path_sign_out: str
) -> dict[str, list[str]]:
    """Create dict of redirect URIs for AppClient."""
    return {
        "sign_in": [f"{domain}{redirect_path_sign_in}" for domain in domains],
        "sign_out": [f"{domain}{redirect_path_sign_out}" for domain in domains],
    }


def update(context: CfnginContext, *_args: Any, **kwargs: Any) -> bool:
    """Update the callback urls for the User Pool Client.

    Required to match the redirect_uri being sent which contains
    our distribution and alternate domain names.

    Arguments parsed by
    :class:`~runway.cfngin.hooks.staticsite.auth_at_edge.client_updater.HookArgs`.

    Args:
        context: The context instance.
        **kwargs: Arbitrary keyword arguments.

    """
    args = HookArgs.model_validate(kwargs)
    session = context.get_session()
    cognito_client = session.client("cognito-idp")

    # Combine alternate domains with main distribution
    redirect_domains = [*args.alternate_domains, "https://" + args.distribution_domain]

    # Create a list of all domains with their redirect paths
    redirect_uris = get_redirect_uris(
        redirect_domains, args.redirect_path_sign_in, args.redirect_path_sign_out
    )
    # Update the user pool client
    try:
        cognito_client.update_user_pool_client(
            AllowedOAuthScopes=args.oauth_scopes,
            AllowedOAuthFlows=["code"],
            SupportedIdentityProviders=args.supported_identity_providers,
            AllowedOAuthFlowsUserPoolClient=True,
            ClientId=args.client_id,
            CallbackURLs=redirect_uris["sign_in"],
            LogoutURLs=redirect_uris["sign_out"],
            UserPoolId=context.hook_data["aae_user_pool_id_retriever"]["id"],
        )
        return True
    except Exception:
        LOGGER.exception("unable to update user pool client callback urls")
        return False



"""File info.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/fileinfo.py

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ......compat import cached_property
from .utils import EPOCH_TIME

if TYPE_CHECKING:
    import datetime

    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import HeadObjectOutputTypeDef, ObjectTypeDef

    from ......type_defs import AnyPath
    from .format_path import SupportedPathType


class FileInfo:
    """Important details related to performing a task.

    It can perform operations such as ``upload``, ``download``, ``copy``,
    ``delete``, ``move``.  Similarly to ``TaskInfo`` objects attributes
    like ``session`` need to be set in order to perform operations.

    """

    def __init__(
        self,
        src: AnyPath,
        *,
        client: S3Client | None = None,
        compare_key: str | None = None,
        dest_type: SupportedPathType | None = None,
        dest: AnyPath | None = None,
        is_stream: bool = False,
        last_update: datetime.datetime | None = None,
        operation_name: str | None = None,
        parameters: dict[str, Any] | None = None,
        response_data: HeadObjectOutputTypeDef | ObjectTypeDef | None = None,
        size: int | None = None,
        source_client: S3Client | None = None,
        src_type: SupportedPathType | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            src: Source path.
            client: boto3 S3 client.
            compare_key: The name of the file relative to the specified
                directory/prefix. This variable is used when performing synching
                or if the destination file is adopting the source file's name.
            dest_type: If the destination is s3 or local.
            dest: Destination path.
            is_stream: If the file is a stream.
            last_update: The local time of last modification.
            operation_name: Name of the operation being run.
            parameters: A dictionary of important values this is assigned in
                the ``BasicTask`` object.
            response_data: The response data used by
                the ``FileGenerator`` to create this task. It is either an dictionary
                from the list of a ListObjects or the response from a HeadObject. It
                will only be filled if the task was generated from an S3 bucket.
            size: The size of the file in bytes.
            source_client: Client to handle the source.
            src_type: If the source is s3 or local.

        """
        self.src = src
        self.src_type = src_type
        self.operation_name = operation_name
        self.client = client
        self.dest = dest
        self.dest_type = dest_type
        self.compare_key = compare_key
        self.size = size
        self.last_update = last_update or EPOCH_TIME
        self.parameters = parameters or {}
        self.source_client = source_client
        self.is_stream = is_stream
        self.associated_response_data = response_data

    @cached_property
    def is_glacier_compatible(self) -> bool:
        """Determine if a file info object is glacier compatible.

        Operations will fail if the S3 object has a storage class of GLACIER
        and it involves copying from S3 to S3, downloading from S3, or moving
        where S3 is the source (the delete will actually succeed, but we do
        not want fail to transfer the file and then successfully delete it).

        Returns:
            True if the FileInfo's operation will not fail because the
            operation is on a glacier object. False if it will fail.

        """
        if self._is_glacier_object(self.associated_response_data):
            if self.operation_name in ["copy", "download"]:
                return False
            if self.operation_name == "move" and self.src_type == "s3":
                return False
        return True

    def _is_glacier_object(
        self, response_data: HeadObjectOutputTypeDef | ObjectTypeDef | None
    ) -> bool:
        """Determine if a file info object is glacier compatible."""
        glacier_storage_classes = ["GLACIER", "DEEP_ARCHIVE"]
        return bool(
            response_data
            and response_data.get("StorageClass") in glacier_storage_classes
            and not self._is_restored(response_data)
        )

    @staticmethod
    def _is_restored(response_data: HeadObjectOutputTypeDef | ObjectTypeDef) -> bool:
        """Return True is this is a glacier object that has been restored back to S3."""
        # 'Restore' looks like: 'ongoing-request="false", expiry-date="..."'
        return 'ongoing-request="false"' in response_data.get("Restore", "")



"""Core Runway components."""

from ._deploy_environment import DeployEnvironment
from ._deployment import Deployment
from ._module import Module
from ._module_path import ModulePath
from ._module_type import RunwayModuleType, RunwayModuleTypeExtensionsTypeDef

__all__ = [
    "DeployEnvironment",
    "Deployment",
    "Module",
    "ModulePath",
    "RunwayModuleType",
    "RunwayModuleTypeExtensionsTypeDef",
]



"""Sync strategies.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/__init__.py

"""

from .base import BaseSync, MissingFileSync, NeverSync, SizeAndLastModifiedSync
from .delete import DeleteSync
from .exact_timestamps import ExactTimestampsSync
from .register import register_sync_strategies
from .size_only import SizeOnlySync

__all__ = [
    "BaseSync",
    "DeleteSync",
    "ExactTimestampsSync",
    "MissingFileSync",
    "NeverSync",
    "register_sync_strategies",
    "SizeAndLastModifiedSync",
    "SizeOnlySync",
]



#!/usr/bin/env python
"""Module with k8s IAM resources."""
import awacs.autoscaling
import awacs.sts
from awacs.aws import Allow, PolicyDocument, Statement
from awacs.helpers.trust import make_simple_assume_policy
from troposphere import Output, iam

from runway.cfngin.blueprints.base import Blueprint

IAM_POLICY_ARN_PREFIX = "arn:aws:iam::aws:policy/"


class Iam(Blueprint):
    """CFNgin blueprint for creating k8s IAM resources."""

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        template = self.template
        template.set_version("2010-09-09")
        template.set_description("Kubernetes IAM policies - V1.0.0")

        # Resources
        nodeinstancerole = template.add_resource(
            iam.Role(
                "NodeInstanceRole",
                AssumeRolePolicyDocument=make_simple_assume_policy("ec2.amazonaws.com"),
                ManagedPolicyArns=[
                    IAM_POLICY_ARN_PREFIX + i
                    for i in [
                        "AmazonEKSWorkerNodePolicy",
                        "AmazonEKS_CNI_Policy",
                        "AmazonEC2ContainerRegistryReadOnly",
                        "AmazonSSMManagedInstanceCore",
                    ]
                ],
            )
        )
        template.add_output(
            Output(
                "NodeInstanceRole",
                Description="The node instance role name",
                Value=nodeinstancerole.ref(),
            )
        )
        template.add_output(
            Output(
                "NodeInstanceRoleArn",
                Description="The node instance role ARN",
                Value=nodeinstancerole.get_att("Arn"),
            )
        )

        nodeinstanceprofile = template.add_resource(
            iam.InstanceProfile("NodeInstanceProfile", Path="/", Roles=[nodeinstancerole.ref()])
        )
        template.add_output(
            Output(
                "NodeInstanceProfile",
                Description="The node instance profile",
                Value=nodeinstanceprofile.ref(),
            )
        )
        template.add_output(
            Output(
                "NodeInstanceProfileArn",
                Description="The node instance profile ARN",
                Value=nodeinstanceprofile.get_att("Arn"),
            )
        )

        template.add_resource(
            iam.Role(
                "ClusterAutoScalerInstanceRole",
                AssumeRolePolicyDocument=make_simple_assume_policy("ec2.amazonaws.com"),
                Policies=[
                    iam.Policy(
                        PolicyName="cluster-autoscaler",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[
                                        awacs.autoscaling.DescribeAutoScalingGroups,
                                        awacs.autoscaling.DescribeAutoScalingInstances,
                                        awacs.autoscaling.DescribeTags,
                                        awacs.autoscaling.SetDesiredCapacity,
                                        awacs.autoscaling.TerminateInstanceInAutoScalingGroup,
                                    ],
                                    Effect=Allow,
                                    Resource=["*"],
                                )
                            ],
                        ),
                    )
                ],
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.context import CfnginContext

    print(Iam("test", CfnginContext(parameters={"namespace": "test"})).to_json())  # noqa: T201



"""``runway gen-sample`` command group."""

# docs: file://./../../../../docs/source/commands.rst
from typing import Any

import click

from ... import options
from ._cdk_csharp import cdk_csharp
from ._cdk_py import cdk_py
from ._cdk_tsc import cdk_tsc
from ._cfn import cfn
from ._cfngin import cfngin
from ._k8s_cfn_repo import k8s_cfn_repo
from ._k8s_flux_repo import k8s_flux_repo
from ._k8s_tf_repo import k8s_tf_repo
from ._sls_py import sls_py
from ._sls_tsc import sls_tsc
from ._static_angular import static_angular
from ._static_react import static_react
from ._tf import tf

__all__ = [
    "cdk_csharp",
    "cdk_py",
    "cdk_tsc",
    "cfn",
    "cfngin",
    "k8s_cfn_repo",
    "k8s_tf_repo",
    "k8s_flux_repo",
    "sls_py",
    "sls_tsc",
    "static_angular",
    "static_react",
    "tf",
]

COMMANDS = [
    cdk_csharp,
    cdk_py,
    cdk_tsc,
    cfn,
    cfngin,
    k8s_cfn_repo,
    k8s_tf_repo,
    k8s_flux_repo,
    sls_py,
    sls_tsc,
    static_angular,
    static_react,
    tf,
]


@click.group("gen-sample", short_help="generate sample module/project")
@options.debug
@options.no_color
@options.verbose
def gen_sample(**_: Any) -> None:
    """Generate a sample Runway module module/project.

    The sample is created in the current directory.
    If a directory already exists with the name Runway tries to use,
    the sample will not be created.

    """


for cmd in COMMANDS:  # register commands
    gen_sample.add_command(cmd)



"""Base class for AWS responses."""

from http import HTTPStatus
from typing import Any

from pydantic import Field

from ....utils import BaseModel


class ResponseError(BaseModel):
    """Analyse the response from AWS S3 HeadBucket API response.

    Keyword Args:
        Code: A unique short code representing the error that was emitted.
        Message: A longer human readable error message.

    """

    code: str = Field(default="", alias="Code")
    """A unique short code representing the error that was emitted."""

    message: str = Field(default="", alias="Message")
    """A longer human readable error message."""

    def __bool__(self) -> bool:
        """Implement evaluation of instances as a bool."""
        return bool(self.code or self.message)


class ResponseMetadata(BaseModel):
    """Analyse the response from AWS S3 HeadBucket API response.

    Keyword Args:
        HostId: Host ID data.
        HTTPHeaders: A map of response header keys and their respective values.
        HTTPStatusCode: The HTTP status code of the response (e.g., 200, 404).
        RequestId: The unique request ID associated with the response.
            Log this value when debugging requests for AWS support.
        RetryAttempts: The number of retries that were attempted before the
            request was completed.

    """

    host_id: str = Field(default="", alias="HostId")
    """Host ID data."""

    https_headers: dict[str, Any] = Field(default={}, alias="HTTPHeaders")
    """A map of response header keys and their respective values."""

    http_status_code: int = Field(default=200, alias="HTTPStatusCode")
    """he HTTP status code of the response (e.g., 200, 404)."""

    request_id: str = Field(default="", alias="RequestId")
    """The unique request ID associated with the response.
    Log this value when debugging requests for AWS support.

    """

    retry_attempts: int = Field(default=0, alias="RetryAttempts")
    """The number of retries that were attempted before the request was completed."""

    @property
    def forbidden(self) -> bool:
        """Whether the response returned 403 (forbidden)."""
        return self.http_status_code == HTTPStatus.FORBIDDEN

    @property
    def not_found(self) -> bool:
        """Whether the response returned 404 (Not Found)."""
        return self.http_status_code == HTTPStatus.NOT_FOUND


class BaseResponse(BaseModel):
    """Analyse the response from AWS S3 HeadBucket API response.

    Keyword Args:
        Error: Information about a service or networking error.
        ResponseMetadata: Information about the request.

    """

    error: ResponseError = Field(default=ResponseError(), alias="Error")
    """Information about a service or networking error."""

    metadata: ResponseMetadata = Field(default=ResponseMetadata(), alias="ResponseMetadata")
    """Information about the request."""



"""Output CFNgin configuration file schema."""

from __future__ import annotations

import json
import locale
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ....config.models.cfngin import CfnginConfigDefinitionModel
from ... import options

if TYPE_CHECKING:
    from runway._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("cfngin", short_help="config schema")
@options.debug
@click.option(
    "--indent",
    default=4,
    help="Number of spaces to use per indentation level when output JSON.",
    metavar="<int>",
    show_default=True,
    type=click.INT,
)
@options.no_color
@click.option(
    "-o",
    "--output",
    default=None,
    help="If provided, schema will be saved to a file instead of being output to stdout.",
    metavar="<file-name>",
)
@options.verbose
def cfngin(indent: int, output: str | None, **_: Any) -> None:
    """Output JSON schema for CFNgin configuration files.

    The schema that is output can be used to validate configuration files.
    It can also be added to text editors to provide autocompletion, tool tips,
    and suggestions within configuration files.

    """
    content = json.dumps(CfnginConfigDefinitionModel.model_json_schema(), indent=indent)
    if output:
        file_path = Path(output).absolute()
        file_path.write_text(  # append empty line to end of file
            content + "\n", encoding=locale.getpreferredencoding(do_setlocale=False)
        )
        LOGGER.success("output JSON schema to %s", file_path)
    else:
        click.echo(content)



"""Shared functionality for the Auth@Edge Lambda suite."""

import base64
import hmac
import json
import logging
import re
import time
from datetime import datetime
from hashlib import sha256
from random import random
from urllib import request
from urllib.parse import urlencode

LOGGER = logging.getLogger(__name__)


def get_config():
    """Retrieve the configuration variables for the Auth@Edge suite.

    Lambda@Edge restricts the ability to use environment variables.
    This configuration object is generated with hard coded values via Runway.

    """
    # This configuration will be replaced with dynamic values
    # via the lambda_config.py Runway hook. Please review
    # that file for more information.
    config = {
        "client_id": "tbd",
        "cognito_auth_domain": "tbd",
        "cookie_settings": {},
        "http_headers": {},
        "nonce_signing_secret": "tbd",
        "oauth_scopes": [],
        "redirect_path_auth_refresh": "tbd",
        "redirect_path_sign_in": "tbd",
        "redirect_path_sign_out": "tbd",
        "required_group": "tbd",
        "user_pool_id": "tbd",
    }

    user_pool_region = "us-east-1"
    region_match = re.match(r"^(\\S+?)_\\S+$", config["user_pool_id"])
    if region_match:
        user_pool_region = region_match.groups()[0]

    config["cloud_front_headers"] = as_cloud_front_headers(config["http_headers"])
    config["token_issuer"] = (
        f"https://cognito-idp.{user_pool_region}.amazonaws.com/{config['user_pool_id']}"
    )
    config["token_jwks_uri"] = f"{config['token_issuer']}/.well-known/jwks.json"
    return config


def as_cloud_front_headers(headers):
    """Convert a series of headers to CloudFront compliant ones.

    Args:
         headers (dict[str, str]): The request/response headers in
            dictionary format.

    """
    res = {}
    for key, value in headers.items():
        res[key.lower()] = [{"key": key, "value": value}]
    return res


def extract_and_parse_cookies(headers, client_id, cookie_compatibility="amplify"):
    """Extract and parse the Cognito cookies from the headers.

    Args:
        headers (dict[str, str]): The request/response headers in
            dictionary format.
        client_id (str): The Cognito UserPool Client ID.
        cookie_compatibility (str): "amplify" or "elasticsearch".

    """
    cookies = extract_cookies_from_headers(headers)
    if not cookies:
        return {}

    if cookie_compatibility == "amplify":
        cookie_names = get_amplify_cookie_names(client_id, cookies)
    elif cookie_compatibility == "elasticsearch":
        cookie_names = get_elasticsearch_cookie_names()

    return {
        "token_user_name": (
            cookies.get(cookie_names["last_user_key"]) if "last_user_key" in cookie_names else None
        ),
        "id_token": cookies.get(cookie_names["id_token_key"]),
        "access_token": cookies.get(cookie_names["access_token_key"]),
        "refresh_token": cookies.get(cookie_names["refresh_token_key"]),
        "scopes": (cookies.get(cookie_names["scope_key"]) if "scope_key" in cookie_names else None),
        "nonce": cookies.get("spa-auth-edge-nonce"),
        "nonce_hmac": cookies.get("spa-auth-edge-nonce-hmac"),
        "pkce": cookies.get("spa-auth-edge-pkce"),
    }


def extract_cookies_from_headers(headers):
    """Extract all cookies from the response headers.

    Args:
         headers (dict[str, dict[str, str]]): The request/response headers in
            dictionary format.

    """
    if "cookie" not in headers:
        return {}

    cookies = {}
    for header in headers["cookie"]:
        split = header.get("value", "").split(";")
        for part in split:
            seq = part.split("=")
            key, value = seq[0], seq[1:]
            cookies[key.strip()] = "=".join(value).strip()

    return cookies


def decode_token(jwt):
    """Decode the JWT and load it's respective parts as JSON.

    Args:
        jwt (str): The JSON Web Token to parse/decode.

    """
    token_body = jwt.split(".")[1]
    decodable_token_body = re.sub(r"\-", "+", token_body)
    decodable_token_body = re.sub(r"\_", "/", token_body)
    decodable_token_body = base64.b64decode(decodable_token_body + "===")
    return json.loads(decodable_token_body.decode("utf8").replace("'", '"'))


def with_cookie_domain(distribution_domain_name, cookie_settings):
    """Check to see if the cookie has a domain, if not then add an Amplify JS compatible one.

    Args:
        distribution_domain_name (str): The domain name of the
            CloudFront distribution.
        cookie_settings (str): The other settings for the cookie.

    """
    try:
        cookie_settings.lower().index("domain")
        return cookie_settings
    except ValueError:
        # Add leading dot for compatibility with Amplify (js-cookie)
        return f"{cookie_settings}; Domain=.{distribution_domain_name}"


def get_amplify_cookie_names(client_id, cookies_or_username):
    """Return mapping dict for cookie names for amplify."""
    key_prefix = f"CognitoIdentityServiceProvider.{client_id}"
    last_user_key = f"{key_prefix}.LastAuthUser"
    if isinstance(cookies_or_username, str):
        token_user_name = cookies_or_username
    else:
        token_user_name = cookies_or_username.get(last_user_key)
    return {
        "last_user_key": last_user_key,
        "user_data_key": f"{key_prefix}.{token_user_name}.userData",
        "scope_key": f"{key_prefix}.{token_user_name}.tokenScopesString",
        "id_token_key": f"{key_prefix}.{token_user_name}.idToken",
        "access_token_key": f"{key_prefix}.{token_user_name}.accessToken",
        "refresh_token_key": f"{key_prefix}.{token_user_name}.refreshToken",
    }


def get_elasticsearch_cookie_names():
    """Return mapping dict for cookie names for elasticsearch."""
    return {
        "id_token_key": "ID-TOKEN",
        "access_token_key": "ACCESS-TOKEN",
        "refresh_token_key": "REFRESH-TOKEN",
        "cognito_enabled_key": "COGNITO-ENABLED",
    }


def generate_cookie_headers(
    event,
    client_id,
    oauth_scopes,
    tokens,
    domain_name,
    cookie_settings,
    cookie_compatibility="amplify",
):
    """Retrieve all cookie headers for our request.

    Return as CloudFront formatted headers.

    Args:
        event (str): "new_tokens" | "sign_out" | "refresh_failed".
        client_id (str): The Cognito UserPool Client ID.
        oauth_scopes (List): The scopes for oauth validation.
        tokens (dict[str, str]): The tokens received from
            the Cognito Request (id, access, refresh).
        domain_name (str): The Domain name the cookies are
            to be associated with.
        cookie_settings (dict[str, str]): The various settings
            that we would like for the various tokens.
        cookie_compatibility (str): "amplify" | "elasticsearch".

    """
    decoded_id_token = decode_token(tokens["id_token"])
    token_user_name = decoded_id_token.get("cognito:username")

    if cookie_compatibility == "amplify":
        cookie_names = get_amplify_cookie_names(client_id, token_user_name)
        user_data = {
            "UserAttributes": [
                {"Name": "sub", "Value": decoded_id_token.get("sub")},
                {"Name": "email", "Value": decoded_id_token.get("email")},
            ],
            "Username": token_user_name,
        }

        cookies = {
            cookie_names["last_user_key"]: "%s; %s"
            % (
                token_user_name,
                with_cookie_domain(domain_name, cookie_settings.get("idToken")),
            ),
            cookie_names["scope_key"]: "%s; %s"
            % (
                " ".join(oauth_scopes),
                with_cookie_domain(domain_name, cookie_settings.get("accessToken")),
            ),
            cookie_names["user_data_key"]: "%s; %s"
            % (
                urlencode(user_data),
                with_cookie_domain(domain_name, cookie_settings.get("idToken")),
            ),
            "amplify-signin-with-hostedUI": "true; %s"
            % (with_cookie_domain(domain_name, cookie_settings.get("accessToken"))),
        }
    elif cookie_compatibility == "elasticsearch":
        cookie_names = get_elasticsearch_cookie_names()
        cookies = {
            cookie_names["cognito_enabled_key"]: "True; "
            + str(with_cookie_domain(domain_name, cookie_settings.get("cognitoEnabled"))),
        }
    cookies[cookie_names["id_token_key"]] = f"{tokens.get('id_token')}; " + str(
        with_cookie_domain(domain_name, cookie_settings.get("idToken")),
    )
    cookies[cookie_names["access_token_key"]] = f"{tokens.get('access_token')}; " + str(
        with_cookie_domain(domain_name, cookie_settings.get("accessToken")),
    )
    cookies[cookie_names["refresh_token_key"]] = f"{tokens.get('refresh_token')}; " + str(
        with_cookie_domain(domain_name, cookie_settings.get("refreshToken"))
    )
    cookies_iter = cookies  # type: ignore
    if event == "sign_out":
        for key in cookies_iter:
            cookies[key] = expire_cookie(cookies[key])
    elif event == "refresh_failed":
        cookies[cookie_names["refresh_token_key"]] = expire_cookie(
            cookies[cookie_names["refresh_token_key"]]
        )

    # https://github.com/aws-samples/cloudfront-authorization-at-edge/issues/89
    for i in ["spa-auth-edge-nonce", "spa-auth-edge-nonce-hmac", "spa-auth-edge-pkce"]:
        if i in cookies:
            cookies[i] = expire_cookie(cookies[i])

    # Return cookies in the form of CF headers
    return [{"key": "set-cookie", "value": f"{key}={val}"} for key, val in cookies.items()]


def expire_cookie_filter(cookie):
    """Filter to determine if a cookie starts with 'max-age' or 'expires'.

    Args:
        cookie (str): The cookie to filter.

    """
    if cookie.lower().startswith("max-age") or cookie.lower().startswith("expires"):
        return False
    return True


def expire_cookie(cookie):
    """Add an expiration property to a specific cookie.

    Args:
        cookie (str): The cookie string.

    """
    cookie_parts = cookie.split(";")
    mapped = [c.strip() for c in cookie_parts]
    filtered = filter(expire_cookie_filter, mapped)
    listed = list(filtered)
    _, settings = listed[0], listed[1:]
    settings.insert(0, "")
    settings.append(f"Expires={datetime.utcnow()}")
    return "; ".join(settings)


def http_post_with_retry(url, data, headers):
    """Given a URL/Data/Headers make a POST request with exponential backoff.

    Used for retrieving token information from Cognito

    Args:
        url (str): The URL to make the POST request to.
        data (dict[str, str]): The dictionary of data elements to
            send with the request (urlencoded internally).
        headers (dict[str, str]): Any headers to send with
            the POST request.

    """
    attempts = 1
    while attempts:
        try:
            encoded_data = urlencode(data).encode()
            req = request.Request(url, encoded_data, headers)
            with request.urlopen(req) as opened_url:
                res = opened_url.read()
            read = res.decode("utf-8")
            json_data = json.loads(read)
            return json_data
        except Exception as err:
            LOGGER.error("HTTP POST to %s failed (attempt %s)", url, attempts)
            LOGGER.error(err)
            if attempts >= 5:
                raise
            if attempts >= 2:
                # After attempting twice do some exponential backoff with jitter
                time.sleep((25 * (pow(2, attempts) + random() * attempts)) / 1000)
        finally:
            attempts += 1


def create_error_html(title, message, link_uri, link_text):
    """Create a basic error html page for exception returns.

    Args:
        title (str): The title of the page.
        message (str): Any exception message.
        link_uri (str): Link url.
        link_text (str): Link displayed text.

    """
    return (
        "<!DOCTYPE html>"
        + "<html lang='en'>"
        + "    <head>"
        + "        <meta charset='utf-8'>"
        + f"        <title>{title}</title>"
        + "    </head>"
        + "    <body>"
        + f"        <h1>{title}</h1>"
        + f"        <p>{message}</p>"
        + f"        <a href='{link_uri}'>{link_text}</a>"
        + "    </body>"
        + "</html>"
    )


def timestamp_in_seconds():
    """Return int of current unix time."""
    return round(time.time())


def sign(string_to_sign, secret, signature_length=16):
    """Create HMAC signature for string."""
    hashed = hmac.new(bytes(secret, "ascii"), bytes(string_to_sign, "ascii"), sha256)
    return base64.urlsafe_b64encode(hashed.digest()).decode()[0:signature_length]



"""Type definitions."""

from __future__ import annotations

from typing_extensions import TypedDict


class AwsLambdaHookDeployResponseTypedDict(TypedDict):
    """Dict output of :class:`runway.cfngin.hooks.awslambda.models.response.AwsLambdaHookDeployResponse` using aliases."""  # noqa: E501

    CodeSha256: str
    Runtime: str
    S3Bucket: str
    S3Key: str
    S3ObjectVersion: str | None



"""``runway whichenv`` command."""

# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any

import click

from ...utils import SafeHaven
from .. import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("whichenv", short_help="current deploy environment")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def whichenv(ctx: click.Context, **_: Any) -> None:
    """Print the current deploy environment name to stdout.

    When run, the deploy environment will be determined from one of the
    following (in order of precedence):

    \b
      - "DEPLOY_ENVIRONMENT" environment variable
      - git branch name (strips "ENV-" prefix, master => common)
      - current working directory

    """  # noqa: D301
    if not (ctx.obj.debug or ctx.obj.verbose):
        logging.getLogger("runway").setLevel(logging.ERROR)  # suppress warnings
    with SafeHaven(environ={"CI": "1"}):  # prevent prompts
        click.echo(ctx.obj.env.name)



"""``runway gen-sample cfngin`` command."""

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import ROOT, TEMPLATES, copy_sample

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("cfngin", short_help="cfngin + troposphere (sampleapp.cfn)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def cfngin(ctx: click.Context, **_: Any) -> None:
    """Generate a sample CFNgin project using Blueprints."""
    src = TEMPLATES / "cfngin"
    src_blueprints = ROOT / "blueprints"
    dest = Path.cwd() / "sampleapp.cfn"
    blueprints = dest / "tfstate_blueprints"
    tf_state = blueprints / "tf_state.py"

    copy_sample(ctx, src, dest)

    blueprints.mkdir()
    LOGGER.debug('copying blueprint from "%s" to "%s"', src_blueprints, blueprints)
    shutil.copyfile(src_blueprints / "__init__.py", blueprints / "__init__.py")
    shutil.copyfile(src_blueprints / "tf_state.py", tf_state)
    tf_state.chmod(tf_state.stat().st_mode | 0o0111)
    LOGGER.success("Sample CFNgin module created at %s", dest)



"""Type definitions."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

AnyPath: TypeAlias = "Path | str"
AnyPathConstrained = TypeVar("AnyPathConstrained", Path, str)


class Boto3CredentialsTypeDef(TypedDict, total=False):
    """Boto3 credentials."""

    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str


class EnvVarsAwsCredentialsTypeDef(TypedDict, total=False):
    """AWS credentials from/for environment variables."""

    AWS_ACCESS_KEY_ID: str
    AWS_SECRET_ACCESS_KEY: str
    AWS_SESSION_TOKEN: str



"""Serverless Hello World function."""

from __future__ import annotations

import json
from typing import Any


def handler(event: Any, context: Any) -> dict[str, int | str]:  # noqa: ARG001
    """Return Serverless Hello World."""
    body = {
        "message": "Go Serverless v1.0! Your function executed successfully!",
        "input": event,
    }
    return {"statusCode": 200, "body": json.dumps(body)}

    # Use this code if you don't use the http event with the LAMBDA-PROXY
    # integration
    # return {
    #     "message": "Go Serverless v1.0! Your function executed!",
    #     "event": event
    # }



"""CDK module."""

from __future__ import annotations

import logging
import platform
import subprocess
import sys
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Literal

from .._logging import PrefixAdaptor
from ..compat import cached_property
from ..config.models.runway.options.cdk import RunwayCdkModuleOptionsDataModel
from ..utils import fix_windows_command_list
from .base import ModuleOptions, RunwayModuleNpm
from .utils import generate_node_command, run_module_command

if TYPE_CHECKING:
    from pathlib import Path

    from .._logging import RunwayLogger
    from ..context import RunwayContext

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))

CdkCommandTypeDef = Literal[
    "bootstrap",
    "context",
    "deploy",
    "destroy",
    "diff",
    "docs",
    "doctor",
    "init",
    "list",
    "metadata",
    "synthesize",
]


class CloudDevelopmentKitOptions(ModuleOptions):
    """Module options for AWS Cloud Development Kit.

    Attributes:
        build_steps: A list of commands to be executed before each action (e.g.
            diff, deploy, destroy).
        data: Options parsed into a data model.
        skip_npm_ci: Skip running ``npm ci`` in the module directory prior to
            processing the module.

    """

    def __init__(self, data: RunwayCdkModuleOptionsDataModel) -> None:
        """Instantiate class.

        Args:
            data: Options parsed into a data model.

        """
        self.build_steps = data.build_steps
        self.data = data
        self.skip_npm_ci = data.skip_npm_ci

    @classmethod
    def parse_obj(cls, obj: object) -> CloudDevelopmentKitOptions:
        """Parse options definition and return an options object.

        Args:
            obj: Object to parse.

        """
        return cls(data=RunwayCdkModuleOptionsDataModel.model_validate(obj))


class CloudDevelopmentKit(RunwayModuleNpm[CloudDevelopmentKitOptions]):
    """CDK Runway Module."""

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: bool | None = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: str | None = None,
        options: dict[str, Any] | ModuleOptions | None = None,
        parameters: dict[str, Any] | None = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=CloudDevelopmentKitOptions.parse_obj(options or {}),
            parameters=parameters,
        )
        # logger needs to be created here to use the correct logger
        self.logger = PrefixAdaptor(self.name, LOGGER)

    @cached_property
    def cli_args(self) -> list[str]:
        """Generate CLI args from self used in all CDK commands."""
        result: list[str] = []
        if self.ctx.no_color:
            result.append("--no-color")
        if self.ctx.env.debug:
            result.append("--debug")
        elif self.ctx.env.verbose:
            result.append("--verbose")
        return result

    @cached_property
    def cli_args_context(self) -> list[str]:
        """Generate CLI args from self passed to CDK commands as ``--context``."""
        result: list[str] = []
        args = {"environment": self.ctx.env.name}
        args.update(self.parameters)
        for key, val in args.items():
            result.extend(["--context", f"{key}={val}"])
        return result

    @cached_property
    def skip(self) -> bool:
        """Determine if the module should be skipped."""
        if self.package_json_missing():
            self.logger.info(
                'skipped; package.json with "aws-cdk" in dependencies or '
                "devDependencies is required for this module type"
            )
            return True
        if not self.explicitly_enabled:
            self.logger.info("skipped; environment required but not defined")
            return True
        return False

    def cdk_bootstrap(self) -> None:
        """Execute ``cdk bootstrap`` command."""
        self.logger.info("init (in progress)")
        run_module_command(
            cmd_list=self.gen_cmd("bootstrap", include_context=True),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.info("init (complete)")

    def cdk_deploy(self) -> None:
        """Execute ``cdk deploy`` command."""
        self.logger.info("deploy (in progress)")
        run_module_command(
            cmd_list=self.gen_cmd("deploy", ['"*"'], include_context=True),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.info("deploy (complete)")

    def cdk_destroy(self) -> None:
        """Execute ``cdk destroy`` command."""
        self.logger.info("destroy (in progress)")
        run_module_command(
            cmd_list=self.gen_cmd("destroy", ['"*"'], include_context=True),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.info("destroy (complete)")

    def cdk_diff(self, stack_name: str | None = None) -> None:
        """Execute ``cdk diff`` command."""
        self.logger.info("plan (in progress)")
        try:
            run_module_command(
                cmd_list=self.gen_cmd(
                    "diff",
                    args_list=[stack_name] if stack_name else None,
                    include_context=True,
                ),
                env_vars=self.ctx.env.vars,
                exit_on_error=False,
                logger=self.logger,
            )
        except subprocess.CalledProcessError as exc:
            self.logger.error("CDK returned %s when running diff", exc.returncode)
            self.logger.error(
                "this can be the result of a runtime error or the stack (%s) "
                "differing from what has been deployed if aws-cdk:enableDiffNoFail "
                "is not enabled",
                stack_name,
            )
            # TODO (kyle): raise error instead of sys.exit() when refactoring cli error handling
            sys.exit(exc.returncode)
        self.logger.info("plan (complete)")

    def cdk_list(self) -> list[str]:
        """Execute ``cdk list`` command."""
        result = subprocess.check_output(
            self.gen_cmd("list", include_context=True),
            env=self.ctx.env.vars,
        ).decode()
        result = result.strip().split("\n")
        LOGGER.debug("found stacks: %s", result)
        return result

    def deploy(self) -> None:
        """Run cdk deploy."""
        if self.skip:
            return
        self.npm_install()
        self.run_build_steps()
        self.cdk_bootstrap()
        self.cdk_deploy()

    def destroy(self) -> None:
        """Run cdk destroy."""
        if self.skip:
            return
        self.npm_install()
        self.run_build_steps()
        self.cdk_destroy()

    def gen_cmd(
        self,
        command: CdkCommandTypeDef,
        args_list: list[str] | None = None,
        *,
        include_context: bool = False,
    ) -> list[str]:
        """Generate and log a CDK command.

        This does not execute the command, only prepares it for use.

        Args:
            command: The CDK command to be executed.
            args_list: Additional arguments to include in the generated command.
            include_context: Optionally, pass context to the CLI. Context is not
                valid for all commands.

        Returns:
            The full command to be passed into a subprocess.

        """
        args = [command, *self.cli_args]
        args.extend(args_list or [])
        if include_context:
            args.extend(self.cli_args_context)
        if self.ctx.env.ci:  # append options that remove interaction
            if command == "deploy":
                args.extend(["--ci", "--require-approval=never"])
            if command == "destroy":
                args.append("--force")
        return generate_node_command(
            command="cdk",
            command_opts=args,
            logger=self.logger,
            package="aws-cdk",
            path=self.path,
        )

    def init(self) -> None:
        """Run cdk bootstrap."""
        if self.skip:
            return
        self.npm_install()
        self.run_build_steps()
        self.cdk_bootstrap()

    def plan(self) -> None:
        """Run cdk diff."""
        if self.skip:
            return
        self.npm_install()
        self.run_build_steps()
        for stack_name in self.cdk_list():
            self.cdk_diff(stack_name)

    def run_build_steps(self) -> None:
        """Run build steps."""
        if not self.options.build_steps:
            return
        self.logger.info("build steps (in progress)")
        for step in self.options.build_steps:
            cmd_list = step.split(" ")
            if platform.system() == "Windows":
                cmd_list = fix_windows_command_list(cmd_list)
            try:
                subprocess.check_call(cmd_list, env=self.ctx.env.vars, cwd=self.path)
            except FileNotFoundError:
                self.logger.error(
                    'attempted to run "%s" but failed to find it (are you sure it '
                    "is installed and available in your PATH?)"
                )
                raise
        self.logger.info("build steps (complete)")



"""CFNgin base action."""

from __future__ import annotations

import logging
import os
import sys
import threading
from typing import TYPE_CHECKING, Any, Callable, ClassVar

import botocore.exceptions

from ..dag import ThreadedWalker, UnlimitedSemaphore, walk
from ..exceptions import CfnginBucketNotFound, PlanFailed
from ..plan import Graph, Plan, Step, merge_graphs
from ..utils import ensure_s3_bucket, get_s3_endpoint, stack_template_key_name

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

    from ...context import CfnginContext
    from ..blueprints.base import Blueprint
    from ..providers.aws.default import Provider, ProviderBuilder
    from ..stack import Stack

LOGGER = logging.getLogger(__name__)

# After submitting a stack update/create, this controls how long we'll wait
# between calls to DescribeStacks to check on it's status. Most stack updates
# take at least a couple minutes, so 30 seconds is pretty reasonable and inline
# with the suggested value in
# https://github.com/boto/botocore/blob/1.6.1/botocore/data/cloudformation/2010-05-15/waiters-2.json#L22
#
# This can be controlled via an environment variable, mostly for testing.
STACK_POLL_TIME = int(os.environ.get("CFNGIN_STACK_POLL_TIME", 30))


def build_walker(concurrency: int) -> Callable[..., Any]:
    """Return a function for waling a graph.

    Passed to :class:`runway.cfngin.plan.Plan` for walking the graph.

    If concurrency is 1 (no parallelism) this will return a simple topological
    walker that doesn't use any multithreading.

    If concurrency is 0, this will return a walker that will walk the graph as
    fast as the graph topology allows.

    If concurrency is greater than 1, it will return a walker that will only
    execute a maximum of concurrency steps at any given time.

    Args:
        concurrency: Number of threads to use while walking.

    Returns:
        Function to walk a :class:`runway.cfngin.dag.DAG`.

    """
    if concurrency == 1:
        return walk

    semaphore = UnlimitedSemaphore()
    if concurrency > 1:
        semaphore = threading.Semaphore(concurrency)

    return ThreadedWalker(semaphore).walk


def stack_template_url(bucket_name: str, blueprint: Blueprint, endpoint: str) -> str:
    """Produce an s3 url for a given blueprint.

    Args:
        bucket_name: The name of the S3 bucket where the resulting
            templates are stored.
        blueprint: The blueprint object to create the URL to.
        endpoint: The s3 endpoint used for the bucket.

    """
    return f"{endpoint}/{bucket_name}/{stack_template_key_name(blueprint)}"


class BaseAction:
    """Actions perform the actual work of each Command.

    Each action is responsible for building the :class:`runway.cfngin.plan.Plan`
    that will be executed.

    Attributes:
        DESCRIPTION: Description used when creating a plan for an action.
        NAME: Name of the action.
        bucket_name: S3 bucket used by the action.
        bucket_region: AWS region where S3 bucket is located.
        cancel: Cancel handler.
        context: The context for the current run.
        provider_builder: An object that will build a provider that will be
            interacted with in order to perform the necessary actions.
        s3_conn: Boto3 S3 client.

    """

    DESCRIPTION: ClassVar[str] = "Base action"
    NAME: ClassVar[str | None] = None

    bucket_name: str | None
    bucket_region: str | None
    cancel: threading.Event
    context: CfnginContext
    provider_builder: ProviderBuilder | None
    s3_conn: S3Client

    def __init__(
        self,
        context: CfnginContext,
        provider_builder: ProviderBuilder | None = None,
        cancel: threading.Event | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            context: The context for the current run.
            provider_builder: An object that will build a provider that will be
                interacted with in order to perform the necessary actions.
            cancel: Cancel handler.

        """
        self.context = context
        self.provider_builder = provider_builder
        self.bucket_name = context.bucket_name
        self.cancel = cancel or threading.Event()
        self.bucket_region = context.config.cfngin_bucket_region
        if not self.bucket_region and provider_builder:
            self.bucket_region = provider_builder.region
        self.s3_conn = self.context.s3_client

    @property
    def _stack_action(self) -> Callable[..., Any]:
        """Run against a step."""
        raise NotImplementedError

    @property
    def provider(self) -> Provider:
        """Return a generic provider using the default region.

        Used for running things like hooks.

        """
        if not self.provider_builder:
            raise ValueError("ProviderBuilder required to build a provider")
        return self.provider_builder.build()

    def build_provider(self) -> Provider:
        """Build a CFNgin provider."""
        if not self.provider_builder:
            raise ValueError("ProviderBuilder required to build a provider")
        return self.provider_builder.build()

    def ensure_cfn_bucket(self) -> None:
        """CloudFormation bucket where templates will be stored."""
        if self.bucket_name:
            try:
                ensure_s3_bucket(self.s3_conn, self.bucket_name, self.bucket_region, create=False)
            except botocore.exceptions.ClientError:
                raise CfnginBucketNotFound(bucket_name=self.bucket_name) from None

    def execute(self, **kwargs: Any) -> None:
        """Run the action with pre and post steps."""
        try:
            self.pre_run(**kwargs)
            self.run(**kwargs)
            self.post_run(**kwargs)
        except PlanFailed as err:
            LOGGER.error(str(err))
            sys.exit(1)

    def pre_run(self, *, dump: bool | str = False, outline: bool = False, **__kwargs: Any) -> None:
        """Perform steps before running the action."""

    def post_run(self, *, dump: bool | str = False, outline: bool = False, **__kwargs: Any) -> None:
        """Perform steps after running the action."""

    def run(
        self,
        *,
        concurrency: int = 0,
        dump: bool | str = False,
        force: bool = False,
        outline: bool = False,
        tail: bool = False,
        upload_disabled: bool = False,
        **_kwargs: Any,
    ) -> None:
        """Abstract method for running the action."""
        raise NotImplementedError('Subclass must implement "run" method')

    def s3_stack_push(self, blueprint: Blueprint, force: bool = False) -> str:
        """Push the rendered blueprint's template to S3.

        Verifies that the template doesn't already exist in S3 before
        pushing.

        Returns:
            URL to the template in S3.

        """
        if not self.bucket_name:
            raise ValueError("bucket_name required")
        key_name = stack_template_key_name(blueprint)
        template_url = self.stack_template_url(blueprint)
        try:
            template_exists = bool(self.s3_conn.head_object(Bucket=self.bucket_name, Key=key_name))
        except botocore.exceptions.ClientError as err:
            if err.response["Error"]["Code"] == "404":
                template_exists = False
            else:
                raise

        if template_exists and not force:
            LOGGER.debug("CloudFormation template already exists: %s", template_url)
            return template_url
        self.s3_conn.put_object(
            Bucket=self.bucket_name,
            Key=key_name,
            Body=blueprint.rendered.encode(),
            ServerSideEncryption="AES256",
            ACL="bucket-owner-full-control",
        )
        LOGGER.debug("blueprint %s pushed to %s", blueprint.name, template_url)
        return template_url

    def stack_template_url(self, blueprint: Blueprint) -> str:
        """S3 URL for CloudFormation template object."""
        if not self.bucket_name:
            raise ValueError("bucket_name required")
        return stack_template_url(self.bucket_name, blueprint, get_s3_endpoint(self.s3_conn))

    def _generate_plan(
        self,
        tail: bool = False,
        reverse: bool = False,
        require_unlocked: bool = True,
        include_persistent_graph: bool = False,
    ) -> Plan:
        """Create a plan for this action.

        Args:
            tail: Whether to tail the stack progress.
            reverse: If True, execute the graph in reverse (useful for destroy actions).
            require_unlocked: If the persistent graph is locked, an error is raised.
            include_persistent_graph: Include the persistent graph
                in the :class:`runway.cfngin.plan.Plan` (if there is one).
                This will handle basic merging of the local and persistent
                graphs if an action does not require more complex logic.

        """
        tail_fn = self._tail_stack if tail else None

        steps = [
            Step(stack, fn=self._stack_action, watch_func=tail_fn) for stack in self.context.stacks
        ]

        graph = Graph.from_steps(steps)

        if include_persistent_graph and self.context.persistent_graph:
            persist_steps = Step.from_persistent_graph(
                self.context.persistent_graph.to_dict(),
                self.context,
                fn=self._stack_action,
                watch_func=tail_fn,
            )
            persist_graph = Graph.from_steps(persist_steps)
            graph = merge_graphs(graph, persist_graph)

        return Plan(
            context=self.context,
            description=self.DESCRIPTION,
            graph=graph,
            reverse=reverse,
            require_unlocked=require_unlocked,
        )

    def _tail_stack(
        self, stack: Stack, cancel: threading.Event, retries: int = 0, **kwargs: Any
    ) -> None:
        """Tail a stack's event stream."""
        provider = self.build_provider()
        return provider.tail_stack(stack, cancel, action=self.NAME, retries=retries, **kwargs)



"""CFNgin package source models."""

from __future__ import annotations

from typing import Annotated, Any

from pydantic import ConfigDict, Field, model_validator

from ..base import ConfigProperty


class GitCfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a git package source definition.

    Package source located in a git repository.

    """

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Information about git repositories that should be included "
            "in the processing of this configuration file."
        },
        title="CFNgin Git Repository Package Source Definition",
        validate_default=True,
        validate_assignment=True,
    )

    branch: Annotated[
        str | None, Field(title="Git Branch", examples=["ENV-dev", "ENV-prod", "master"])
    ] = None
    """Branch name."""

    commit: Annotated[str | None, Field(title="Git Commit Hash")] = None
    """Commit hash."""

    configs: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source "
            "for configuration that should be merged into the current configuration file."
        ),
    ] = []
    """List of CFNgin config paths to execute."""

    paths: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source to add to $PATH."
        ),
    ] = []
    """List of paths to append to ``sys.path``."""

    tag: Annotated[str | None, Field(title="Git Tag", examples=["1.0.0", "v1.0.0"])] = None
    """Git tag."""

    uri: Annotated[
        str, Field(title="Git Repository URI", examples=["git@github.com:rackspace/runway.git"])
    ]
    """Remote git repo URI."""

    @model_validator(mode="before")
    @classmethod
    def _validate_one_ref(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Ensure that only one ref is defined."""
        ref_keys = ["branch", "commit", "tag"]
        count_ref_defs = sum(bool(values.get(i)) for i in ref_keys)
        if count_ref_defs > 1:
            raise ValueError(f"only one of {ref_keys} can be defined")
        return values


class LocalCfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a CFNgin local package source definition.

    Package source located on a local disk.

    """

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Information about local directories that should be "
            "included in the processing of this configuration file."
        },
        title="CFNgin Local Package Source Definition",
        validate_default=True,
        validate_assignment=True,
    )

    configs: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source "
            "for configuration that should be merged into the current configuration file.",
        ),
    ] = []
    """List of CFNgin config paths to execute."""

    paths: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source to add to $PATH."
        ),
    ] = []
    """List of paths to append to ``sys.path``."""

    source: Annotated[
        str,
        Field(
            description="Path relative to the current configuration file that is the "
            "root of the local package source."
        ),
    ]
    """Source."""


class S3CfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a CFNgin S3 package source definition.

    Package source located in AWS S3.

    """

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Information about a AWS S3 objects that should be "
            "downloaded, unzipped, and included in the processing of "
            "this configuration file."
        },
        title="CFNgin S3 Package Source Definition",
        validate_default=True,
        validate_assignment=True,
    )

    bucket: Annotated[str, Field(title="AWS S3 Bucket Name")]
    """AWS S3 bucket name."""

    configs: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source "
            "for configuration that should be merged into the current configuration file.",
        ),
    ] = []
    """List of CFNgin config paths to execute."""

    key: Annotated[str, Field(title="AWS S3 Object Key")]
    """Object key. The object should be a zip file."""

    paths: Annotated[
        list[str],
        Field(
            description="Array of paths relative to the root of the package source to add to $PATH."
        ),
    ] = []
    """List of paths to append to ``sys.path``."""

    requester_pays: Annotated[
        bool,
        Field(
            description="Confirms that the requester knows that they will be charged for the request."
        ),
    ] = False
    """AWS S3 requester pays option."""

    use_latest: Annotated[
        bool,
        Field(description="Update the local copy if the last modified date in AWS S3 changes."),
    ] = True
    """Use the latest version of the object."""


class CfnginPackageSourcesDefinitionModel(ConfigProperty):
    """Model for a CFNgin package sources definition."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Map of additional package sources to include when "
            "processing this configuration file."
        },
        title="CFNgin Package Sources Definition",
        validate_default=True,
        validate_assignment=True,
    )

    git: list[GitCfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin Git Repository Package Source Definitions",
        description="Information about git repositories that should be included "
        "in the processing of this configuration file.",
    )
    """Package source located in a git repo."""

    local: list[LocalCfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin Local Package Source Definitions",
        description="Information about local directories that should be included "
        "in the processing of this configuration file.",
    )
    """Package source located on a local disk."""

    s3: list[S3CfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin S3 Package Source Definitions",
        description="Information about a AWS S3 objects that should be "
        "downloaded, unzipped, and included in the processing of this configuration file.",
    )
    """Package source located in AWS S3."""



"""Copy of ``awscli.customizations.cloudformation.yamlhelper.py``."""

# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from __future__ import annotations

import json
from collections.abc import MutableMapping, MutableSequence
from typing import Any, cast

import yaml


def intrinsics_multi_constructor(
    loader: yaml.Loader, tag_prefix: str, node: yaml.Node  # noqa: ARG001
) -> dict[str, Any]:
    """YAML constructor to parse CloudFormation intrinsics.

    This will return a dictionary with key being the intrinsic name

    """
    # Get the actual tag name excluding the first exclamation
    tag = node.tag[1:]

    # Some intrinsic functions doesn't support prefix "Fn::"
    prefix = "Fn::"
    if tag in ["Ref", "Condition"]:
        prefix = ""

    cfntag = prefix + tag

    if tag == "GetAtt" and isinstance(node.value, str):
        # ShortHand notation for !GetAtt accepts Resource.Attribute format
        # while the standard notation is to use an array
        # [Resource, Attribute]. Convert shorthand to standard format
        value = node.value.split(".", 1)

    elif isinstance(node, yaml.ScalarNode):
        # Value of this node is scalar
        value = loader.construct_scalar(node)

    elif isinstance(node, yaml.SequenceNode):
        # Value of this node is an array (Ex: [1,2])
        value = cast(MutableSequence[Any], loader.construct_sequence(node))

    else:
        # Value of this node is an mapping (ex: {foo: bar})
        value = cast(MutableMapping[Any, Any], loader.construct_mapping(node))  # type: ignore

    return {cfntag: value}


def yaml_dump(dict_to_dump: dict[str, Any]) -> str:
    """Dump the dictionary as a YAML document."""
    return yaml.safe_dump(dict_to_dump, default_flow_style=False)


def yaml_parse(yamlstr: str) -> dict[str, Any]:
    """Parse a yaml string."""
    try:
        # PyYAML doesn't support json as well as it should, so if the input
        # is actually just json it is better to parse it with the standard
        # json parser.
        return json.loads(yamlstr)
    except ValueError:
        yaml.SafeLoader.add_multi_constructor("!", intrinsics_multi_constructor)
        return yaml.safe_load(yamlstr)



"""Runway exceptions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .utils import DOC_SITE

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from .variables import (
        Variable,
        VariableValue,
        VariableValueConcatenation,
        VariableValueLookup,
    )


class RunwayError(Exception):
    """Base class for custom exceptions raised by Runway."""

    message: str
    """Error message."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class."""
        if self.message:
            super().__init__(self.message, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)


class ConfigNotFound(RunwayError):
    """Configuration file could not be found."""

    looking_for: list[str]
    message: str
    path: Path

    def __init__(self, *, looking_for: list[str] | None = None, path: Path) -> None:
        """Instantiate class.

        Args:
            path: Path where the config file was expected to be found.
            looking_for: List of file names that were being looked for.

        """
        self.looking_for = looking_for or []
        self.path = path

        if looking_for:
            self.message = f"config file not found at path {path}; looking for one of {looking_for}"
        else:
            self.message = f"config file not found at path {path}"
        super().__init__(self.path, self.looking_for)


class DockerConnectionRefusedError(RunwayError):
    """Docker connection refused.

    This can be caused by a number of reasons:

    - Docker is not installed.
    - Docker service is not running.
    - The current user does not have adequate permissions (e.g. not a member of
      the ``docker`` group).

    """

    def __init__(self) -> None:
        """Instantiate class."""
        self.message = (
            "Docker connection refused; install Docker, ensure it is running, "
            "and ensure the current user has adequate permissions"
        )
        super().__init__()


class DockerExecFailedError(RunwayError):
    """Docker failed when trying to execute a command.

    This can be used for ``docker container run``, ``docker container start``
    and ``docker exec`` equivalents.

    """

    exit_code: int
    """The ``StatusCode`` returned by Docker."""

    def __init__(self, response: dict[str, Any]) -> None:
        """Instantiate class.

        Args:
            response: The return value of :meth:`docker.models.containers.Container.wait`,
                Docker API's response as a Python dictionary.
                This can contain important log information pertinent to troubleshooting
                that may not streamed.

        """
        self.exit_code = response.get("StatusCode", 1)  # we can assume this will be > 0
        error: dict[Any, Any] = response.get("Error") or {}  # value from dict could be NoneType
        self.message = error.get("Message", "error message undefined")
        super().__init__()


class FailedLookup(RunwayError):
    """Intermediary Exception to be converted to FailedVariableLookup.

    Should be caught by error handling and
    :class:`runway.cfngin.exceptions.FailedVariableLookup` raised instead to
    construct a proper error message.

    """

    cause: Exception
    lookup: VariableValueLookup
    message: str = "Failed lookup"

    def __init__(
        self, lookup: VariableValueLookup, cause: Exception, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            lookup: The variable value lookup that was attempted and
                resulted in an exception being raised.
            cause: The exception that was raised.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.cause = cause
        self.lookup = lookup
        super().__init__(*args, **kwargs)


class FailedVariableLookup(RunwayError):
    """Lookup could not be resolved.

    Raised when an exception is raised when trying to resolve a lookup.

    """

    cause: FailedLookup
    variable: Variable
    message: str

    def __init__(
        self, variable: Variable, lookup_error: FailedLookup, *args: Any, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            variable: The variable containing the failed lookup.
            lookup_error: The exception that was raised directly before this one.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.cause = lookup_error
        self.variable = variable
        self.message = (
            f'Could not resolve lookup "{lookup_error.lookup}" for variable "{variable.name}"'
        )
        super().__init__(*args, **kwargs)


class HclParserError(RunwayError):
    """HCL/HCL2 parser error."""

    message: str

    def __init__(
        self,
        exc: Exception,
        file_path: Path | str,
        parser: ModuleType | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            exc: Exception that was raised.
            file_path: File that resulted in the error.
            parser: The parser what was used to try to parse the file (hcl|hcl2).

        """
        self.reason = exc
        self.file_path = file_path
        if parser:
            self.message = f"Unable to parse {file_path} as {parser.__name__.upper()}\n\n{exc}"
        else:
            self.message = f"Unable to parse {file_path}\n\n{exc}"
        super().__init__()


class InvalidLookupConcatenation(RunwayError):
    """Invalid return value for a concatenated (chained) lookup.

    The return value must be a string when lookups are concatenated.

    """

    concatenated_lookups: VariableValueConcatenation[Any]
    invalid_lookup: VariableValue
    message: str

    def __init__(
        self,
        invalid_lookup: VariableValue,
        concat_lookups: VariableValueConcatenation[Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Instantiate class."""
        self.concatenated_lookups = concat_lookups
        self.invalid_lookup = invalid_lookup
        self.message = (
            f"expected return value of type {str} but received "
            f'{type(invalid_lookup.value)} for lookup "{invalid_lookup}" '
            f'in "{concat_lookups}"'
        )
        super().__init__(*args, **kwargs)


class KubectlVersionNotSpecified(RunwayError):
    """kubectl version is required but was not specified.

    Version can be specified by using a file or option.

    """

    message: str

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class."""
        self.message = (
            "kubectl version not specified. Learn how to use Runway to manage kubectl versions "
            f"at {DOC_SITE}/page/kubernetes/advanced_features.html#version-management"
        )
        super().__init__(*args, **kwargs)


class NpmNotFound(RunwayError):
    """Raised when npm could not be executed or was not found in path."""

    message: str

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class."""
        self.message = (
            '"npm" not found in path or is not executable; '
            "please ensure it is installed correctly"
        )
        super().__init__(*args, **kwargs)


class OutputDoesNotExist(RunwayError):
    """Raised when a specific stack output does not exist."""

    output: str
    """Name of the CloudFormation Stack's Output that does not exist."""

    stack_name: str
    """Name of a CloudFormation Stack."""

    def __init__(self, stack_name: str, output: str, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            stack_name: Name of the stack.
            output: The output that does not exist.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.stack_name = stack_name
        self.output = output

        self.message = f"Output {output} does not exist on stack {stack_name}"
        super().__init__(*args, **kwargs)


class RequiredTagNotFoundError(RunwayError):
    """Required tag not found on resource."""

    resource: str
    """An ID or name to identify a resource."""

    tag_key: str
    """Key of the tag that could not be found."""

    def __init__(self, resource: str, tag_key: str) -> None:
        """Instantiate class.

        Args:
            resource: An ID or name to identify a resource.
            tag_key: Key of the tag that could not be found.

        """
        self.resource = resource
        self.tag_key = tag_key
        self.message = f"required tag '{tag_key}' not found for {resource}"
        super().__init__()


class UnknownLookupType(RunwayError):
    """Lookup type provided does not match a registered lookup.

    Example:
        If a lookup of ``${<lookup_type> query}`` is used and ``<lookup_type>``
        is not a registered lookup, this exception will be raised.

    """

    message: str

    def __init__(self, lookup: VariableValueLookup, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            lookup: Variable value lookup that could not find a handler.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.message = f'Unknown lookup type "{lookup.lookup_name.value}" in "{lookup}"'
        super().__init__(*args, **kwargs)


class UnresolvedVariable(RunwayError):
    """Raised when trying to use a variable before it has been resolved."""

    message: str

    def __init__(self, variable: Variable, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            variable: The unresolved variable.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.message = f'Attempted to use variable "{variable.name}" before it was resolved'
        self.variable = variable
        super().__init__(*args, **kwargs)


class UnresolvedVariableValue(RunwayError):
    """Intermediary Exception to be converted to UnresolvedVariable.

    Should be caught by error handling and
    :class:`runway.cfngin.exceptions.UnresolvedVariable` raised instead to
    construct a proper error message.

    """

    lookup: VariableValueLookup
    message: str = "Unresolved lookup"

    def __init__(self, lookup: VariableValueLookup, *args: Any, **kwargs: Any) -> None:
        """Instantiate class.

        Args:
            lookup: The variable value lookup that is not resolved.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        self.lookup = lookup
        super().__init__(*args, **kwargs)


class VariablesFileNotFound(RunwayError):
    """Defined variables file could not be found."""

    file_path: Path
    message: str

    def __init__(self, file_path: Path) -> None:
        """Instantiate class.

        Args:
            file_path: Path where the file was expected to be found.

        """
        self.file_path = file_path
        self.message = f"defined variables file not found at path {file_path}"
        super().__init__(self.file_path)
